SERIALIZERS = {}
JS_CLASS_MAPPING = {}


def register_instance_serializer(name, method):
    SERIALIZERS[name] = method


def register_js_class(vtk_class, js_class):
    JS_CLASS_MAPPING[vtk_class] = js_class


def class_name(vtk_obj):
    vtk_class = vtk_obj.GetClassName()
    if vtk_class in JS_CLASS_MAPPING:
        return JS_CLASS_MAPPING[vtk_class]

    return vtk_class



import base64

import numpy as np
import os
import importlib
import sys

vtk_module_name = os.environ.get("VTK_MODULE_NAME", "vtkmodules")
sys.modules["vtk_module"] = importlib.import_module(vtk_module_name)

from vtk_module.util.numpy_support import vtk_to_numpy
from vtk_module.vtkFiltersGeometry import vtkDataSetSurfaceFilter


def mesh(dataset, field_to_keep=None, point_arrays=None, cell_arrays=None):
    """Expect any dataset and extract its surface into a dash_vtk.Mesh state property"""
    if dataset is None:
        return None

    # Make sure we have a polydata to export
    polydata = None
    if dataset.IsA("vtkPolyData"):
        polydata = dataset
    else:
        extract_skin_filter = vtkDataSetSurfaceFilter()
        extract_skin_filter.SetInputData(dataset)
        extract_skin_filter.Update()
        polydata = extract_skin_filter.GetOutput()

    if polydata.GetPoints() is None:
        return None

    # Extract mesh
    state = {"mesh": {}}

    points = mesh_array(polydata.GetPoints())
    if points:
        state["mesh"]["points"] = points

    verts = mesh_array(polydata.GetVerts())
    if verts:
        state["mesh"]["verts"] = verts

    lines = mesh_array(polydata.GetLines())
    if lines:
        state["mesh"]["lines"] = lines

    polys = mesh_array(polydata.GetPolys())
    if polys:
        state["mesh"]["polys"] = polys

    strips = mesh_array(polydata.GetStrips())
    if strips:
        state["mesh"]["strips"] = strips

    # Scalars
    if field_to_keep is not None:
        field = None
        p_array = polydata.GetPointData().GetArray(field_to_keep)
        c_array = polydata.GetCellData().GetArray(field_to_keep)

        if c_array:
            field = data_array(c_array, location="CellData", name=field_to_keep)

        if p_array:
            field = data_array(p_array, location="PointData", name=field_to_keep)

        if field:
            state.update({"field": field})

    # PointData Fields
    if point_arrays:
        point_data = field_data(polydata.GetPointData(), point_arrays, "PointData")
        if len(point_data):
            state.update({"pointArrays": point_data})

    # CellData Fields
    if cell_arrays:
        cell_data = field_data(polydata.GetCellData(), cell_arrays, "CellData")
        if len(cell_data):
            state.update({"cellArrays": cell_data})

    return state


def mesh_array(array):
    if array:
        return b64_encode_numpy(vtk_to_numpy(array.GetData()))


def data_array(data_array, location="PointData", name=None):
    if data_array:
        data_range = data_array.GetRange(-1)
        nb_comp = data_array.GetNumberOfComponents()
        values = vtk_to_numpy(data_array)
        js_types = to_js_type[str(values.dtype)]
        return {
            "name": name if name else data_array.GetName(),
            "values": b64_encode_numpy(values),
            "numberOfComponents": nb_comp,
            "type": js_types,
            "location": location,
            "dataRange": data_range,
        }


def field_data(field_data, names, location="PointData"):
    fields = []
    for name in names:
        array = field_data.GetArray(name)
        js_array = data_array(array, location, name)
        if js_array:
            fields.append(js_array)

    return fields


def volume(dataset):
    """Expect a vtkImageData and extract its setting for the dash_vtk.Volume state"""
    if dataset is None or not dataset.IsA("vtkImageData"):
        return None

    state = {
        "image": {
            "dimensions": dataset.GetDimensions(),
            "spacing": dataset.GetSpacing(),
            "origin": dataset.GetOrigin(),
        },
    }

    # Capture image orientation if any
    if hasattr(dataset, "GetDirectionMatrix"):
        matrix = dataset.GetDirectionMatrix()
        js_mat = []
        for j in range(3):
            for i in range(3):
                js_mat.append(matrix.GetElement(i, j))

        state["image"]["direction"] = js_mat

    scalars = dataset.GetPointData().GetScalars()
    field = data_array(scalars, location="PointData")
    if field:
        state["field"] = field

    return state


# Numpy to JS TypedArray
to_js_type = {
    "int8": "Int8Array",
    "uint8": "Uint8Array",
    "int16": "Int16Array",
    "uint16": "Uint16Array",
    "int32": "Int32Array",
    "uint32": "Uint32Array",
    "int64": "Int32Array",
    "uint64": "Uint32Array",
    "float32": "Float32Array",
    "float64": "Float64Array",
}


def b64_encode_numpy(obj):
    # Convert 1_d numpy arrays with numeric types to memoryviews with
    # datatype and shape metadata.
    if len(obj) == 0:
        return obj.tolist()

    dtype = obj.dtype
    if dtype.kind == "f":
        return np_encode(obj)
    elif dtype.kind == "b":
        return np_encode(obj, np.uint8)
    elif dtype.kind in ["u", "i"]:
        # Try to see if we can downsize the array
        max_value = np.amax(obj)
        min_value = np.amin(obj)
        signed = min_value < 0
        test_value = max(max_value, -min_value)
        if signed:
            if test_value < np.iinfo(np.int8):
                return np_encode(obj, np.int8)
            if test_value < np.iinfo(np.int16).max:
                return np_encode(obj, np.int16)
            if test_value < np.iinfo(np.int32).max:
                return np_encode(obj, np.int32)
        else:
            if test_value < np.iinfo(np.uint8).max:
                return np_encode(obj, np.uint8)
            if test_value < np.iinfo(np.uint16).max:
                return np_encode(obj, np.uint16)
            if test_value < np.iinfo(np.uint32).max:
                return np_encode(obj, np.uint32)

    # Convert all other numpy arrays to lists
    return obj.tolist()


def np_encode(array, np_type=None):
    if np_type:
        n_array = array.astype(np_type).ravel(order="C")
        return {
            "bvals": base64.b64encode(memoryview(n_array)).decode("utf-8"),
            "dtype": str(n_array.dtype),
            "shape": list(array.shape),
        }
    return {
        "bvals": base64.b64encode(memoryview(array.ravel(order="C"))).decode("utf-8"),
        "dtype": str(array.dtype),
        "shape": list(array.shape),
    }



import warnings
import importlib
import os
import sys

from .core import HybridView
from .serializers import mesh as vtk_mesh

try:
    vtk_module_name = os.environ.get("VTK_MODULE_NAME", "vtkmodules")
    sys.modules["vtk_module"] = importlib.import_module(vtk_module_name)
    HAS_VTK = True
except ImportError:
    warnings.warn("VTK is not installed.")
    HAS_VTK = False

try:
    from vtk_module.vtkWebCore import vtkWebApplication

    HAS_VTK_WEB = True
except ImportError:
    HAS_VTK_WEB = False

IMPROPER_VTK_MSG = """Your build of VTK does not have the proper web modules enabled.
These modules are typically enabled by default with the
`-DVTK_GROUP_ENABLE_Web:STRING=WANT` build flag.

Conda users: This is a known issue with the conda-forge VTK feedstock.
See https://github.com/conda-forge/vtk-feedstock/pull/258
"""


def has_capabilities(*features):
    if not HAS_VTK_WEB:
        raise ImportError(IMPROPER_VTK_MSG)


class Helper:
    def __init__(self, trame_server):
        self._root_protocol = None
        self._trame_server = trame_server
        if HAS_VTK_WEB:
            self._vtk_core = vtkWebApplication()
            self._vtk_core.SetImageEncoding(0)
            self._hybrid_views = {}

            # Link our custom protocols initialization
            trame_server.add_protocol_to_configure(self.configure_protocol)

    def has_capabilities(self, *features):
        has_capabilities(*features)

    def id(self, vtk_obj):
        if not vtk_obj:
            return ""
        return str(self._vtk_core.GetObjectIdMap().GetGlobalId(vtk_obj))

    def object(self, vtk_id):
        return self._vtk_core.GetObjectIdMap().GetVTKObject(int(vtk_id))

    def mesh(self, dataset, field_to_keep=None, point_arrays=None, cell_arrays=None):
        if dataset.IsA("vtkAlgorithm"):
            dataset.Update()
            dataset = dataset.GetOutput()
        return vtk_mesh(
            dataset,
            field_to_keep=field_to_keep,
            point_arrays=point_arrays,
            cell_arrays=cell_arrays,
        )

    def scene(
        self,
        render_window,
        reset_camera=False,
        new_state=False,
        widgets=None,
        orientation_axis=0,
        **kwargs,
    ):
        scene_state = self._trame_server.protocol_call(
            "viewport.geometry.view.get.state",
            self.id(render_window),
            new_state,
            widgets=widgets,
            orientation_axis=orientation_axis,
        )
        if reset_camera:
            scene_state.setdefault("extra", {})["resetCamera"] = 1
        return scene_state

    def export(self, render_window, widgets=None, orientation_axis=0, **kwargs):
        return self._trame_server.protocol_call(
            "viewport.geometry.view.get.export",
            self.id(render_window),
            widgets=widgets,
            orientation_axis=orientation_axis,
        )

    def push_image(self, render_window, reset_camera=False):
        # Disable any double render...
        render_window.GetInteractor().EnableRenderOff()

        if reset_camera:
            self._trame_server.protocol_call(
                "viewport.camera.reset", self.id(render_window)
            )

        return self._trame_server.protocol_call(
            "viewport.image.push", {"view": self.id(render_window)}
        )

    def get_current_image_quality(self, render_window):
        return self._trame_server.protocol_call(
            "viewport.image.push.quality.get", self.id(render_window)
        )

    def set_image_quality(self, render_window, quality, ratio):
        self._trame_server.protocol_call(
            "viewport.image.push.quality",
            self.id(render_window),
            quality,
            ratio,
        )

    def start_animation(self, render_window, fps=30, quality=100, ratio=1):
        self._trame_server.protocol_call("viewport.image.animation.fps.max", fps)
        self.set_image_quality(render_window, quality, ratio)
        self._trame_server.protocol_call(
            "viewport.image.animation.start", self.id(render_window)
        )

    def stop_animation(self, render_window):
        self._trame_server.protocol_call(
            "viewport.image.animation.stop", self.id(render_window)
        )

    def camera(self, render_window):
        camera = render_window.GetRenderers().GetFirstRenderer().GetActiveCamera()
        return {
            "focalPoint": list(camera.GetFocalPoint()),
            "parallelProjection": camera.GetParallelProjection(),
            "parallelScale": camera.GetParallelScale(),
            "position": list(camera.GetPosition()),
            "viewAngle": camera.GetViewAngle(),
            "viewUp": list(camera.GetViewUp()),
            "centerOfRotation": list(camera.GetFocalPoint()),  # no center of rotation
        }

    def set_camera(self, render_window, **kwargs):
        camera = render_window.GetRenderers().GetFirstRenderer().GetActiveCamera()
        key_to_setter = {
            "focalPoint": "SetFocalPoint",
            "parallelProjection": "SetParallelProjection",
            "parallelScale": "SetParallelScale",
            "position": "SetPosition",
            "viewAngle": "SetViewAngle",
            "viewUp": "SetViewUp",
        }
        for key, setter in key_to_setter.items():
            if key in kwargs:
                getattr(camera, setter)(kwargs[key])

    def view(self, view, name="view", **kwargs):
        return self.add_hybrid_view(name, view, **kwargs)

    def configure_protocol(self, protocol):
        self._root_protocol = protocol
        from .protocols import (
            vtkWebLocalRendering,
            vtkWebMouseHandler,
            vtkWebPublishImageDelivery,
            vtkWebViewPort,
        )

        # Initialize vtk application helper
        self._root_protocol.setSharedObject("app", self._vtk_core)

        # Remote rendering - image delivery
        self._root_protocol.registerLinkProtocol(vtkWebMouseHandler())
        self._root_protocol.registerLinkProtocol(vtkWebViewPort())
        self._root_protocol.registerLinkProtocol(
            vtkWebPublishImageDelivery(decode=False)
        )

        # Remote rendering - geometry delivery
        self._root_protocol.registerLinkProtocol(vtkWebLocalRendering())

    def add_hybrid_view(
        self,
        name,
        view,
        mode="local",
        interactive_ratio=1,
        interactive_quality=60,
        still_ratio=1,
        still_quality=98,
        force_replace=False,
        **kwargs,
    ):
        if name in self._hybrid_views:
            if force_replace:
                self._hybrid_views[name].replace_view(view)
            else:
                print(f"A view with name ({name}) is already registered")
                print(" => returning previous one")
            return self._hybrid_views[name]

        view_helper = HybridView(
            self,
            view,
            name,
            mode,
            interactive_ratio,
            interactive_quality,
            still_ratio,
            still_quality,
        )
        self._hybrid_views[name] = view_helper
        return view_helper

    def remove_hybrid_view(self, name):
        if name in self._hybrid_views:
            self._hybrid_views.pop(name)

    def reload_app(self):
        self._hybrid_views = {}


# -----------------------------------------------------------------------------
# Module advanced initialization
# -----------------------------------------------------------------------------

HELPERS_PER_SERVER = {}


def setup(trame_server, **kwargs):
    global HELPERS_PER_SERVER
    if HAS_VTK_WEB:
        HELPERS_PER_SERVER[trame_server.name] = Helper(trame_server)


def get_helper(server):
    return HELPERS_PER_SERVER.get(server.name)



import io
import logging
import time
import zipfile
import os
import importlib
import sys

vtk_module_name = os.environ.get("VTK_MODULE_NAME", "vtkmodules")
sys.modules["vtk_module"] = importlib.import_module(vtk_module_name)

from vtk_module.vtkCommonCore import vtkTypeUInt32Array, vtkFloatArray, vtkDoubleArray

from .utils import base64_encode, wrap_id

logger = logging.getLogger(__name__)

JS_VTK_ARRAY = {
    "Float32Array": vtkFloatArray,
    "Float64Array": vtkDoubleArray,
}


class SynchronizationContext:
    """Convenience class for caching data arrays, storing computed sha sums, keeping
    track of valid actors, etc..
    """

    def __init__(self):
        self.data_array_cache = {}
        self.last_dependencies_mapping = {}
        self.ignore_last_dependencies = False

    def set_ignore_last_dependencies(self, force):
        self.ignore_last_dependencies = force

    def cache_data_array(self, p_md5, data):
        self.data_array_cache[p_md5] = data

    def get_cached_data_array(self, p_md5, binary=False, compression=False):
        cache_obj = self.data_array_cache[p_md5]
        array = cache_obj["array"]
        cache_time = cache_obj["mTime"]
        array_js_datatype = cache_obj.get("dataType")

        if cache_time != array.GetMTime():
            logger.debug(" ***** ERROR: you asked for an old cache key! ***** ")

        if array_js_datatype and array_js_datatype in JS_VTK_ARRAY:
            logger.debug(
                "Convert Array %s to %s", array.GetClassName(), array_js_datatype
            )
            new_array = JS_VTK_ARRAY[array_js_datatype]()
            new_array.DeepCopy(array)
            p_buffer = memoryview(new_array)
        elif array.GetDataType() == 12:
            # IdType need to be converted to Uint32
            array_size = array.GetNumberOfTuples() * array.GetNumberOfComponents()
            new_array = vtkTypeUInt32Array()
            new_array.SetNumberOfTuples(array_size)
            for i in range(array_size):
                new_array.SetValue(
                    i, -1 if array.GetValue(i) < 0 else array.GetValue(i)
                )
            p_buffer = memoryview(new_array)
        else:
            p_buffer = memoryview(array)

        if binary:
            # Convert the vtkUnsignedCharArray into a bytes object, required by
            # Autobahn websockets
            return (
                p_buffer.tobytes()
                if not compression
                else zip_compression(p_md5, p_buffer.tobytes())
            )

        return base64_encode(
            p_buffer if not compression else zip_compression(p_md5, p_buffer.tobytes())
        )

    def check_for_arrays_to_release(self, time_window=20):
        cut_off_time = time.time() - time_window
        shas_to_delete = []
        for sha in self.data_array_cache:
            record = self.data_array_cache[sha]
            array = record["array"]
            count = array.GetReferenceCount()

            if count == 1 and record["ts"] < cut_off_time:
                shas_to_delete.append(sha)

        for sha in shas_to_delete:
            del self.data_array_cache[sha]

    def get_last_dependency_list(self, idstr):
        last_deps = []
        if (
            idstr in self.last_dependencies_mapping
            and not self.ignore_last_dependencies
        ):
            last_deps = self.last_dependencies_mapping[idstr]
        return last_deps

    def set_new_dependency_list(self, idstr, dep_list):
        self.last_dependencies_mapping[idstr] = dep_list

    def build_dependency_call_list(self, idstr, new_list, add_method, remove_method):
        old_list = self.get_last_dependency_list(idstr)

        calls = []
        calls += [[add_method, [wrap_id(x)]] for x in new_list if x not in old_list]
        calls += [[remove_method, [wrap_id(x)]] for x in old_list if x not in new_list]

        self.set_new_dependency_list(idstr, new_list)
        return calls


def zip_compression(name, data):
    with io.BytesIO() as in_memory:
        with zipfile.ZipFile(in_memory, mode="w") as zf:
            zf.writestr("data/%s" % name, data, zipfile.ZIP_DEFLATED)
        in_memory.seek(0)
        return in_memory.read()



import io
import struct
import time

from .utils import array_types_mapping, get_js_array_type, reference_id, hash_data_array


# -----------------------------------------------------------------------------
# Array helpers
# -----------------------------------------------------------------------------


def data_table_to_list(data_table):
    data_type = array_types_mapping[data_table.GetDataType()]
    element_size = struct.calcsize(data_type)
    nb_values = data_table.GetNumberOfValues()
    nb_components = data_table.GetNumberOfComponents()
    nbytes = element_size * nb_values
    if data_type != " ":
        with io.BytesIO(memoryview(data_table)) as stream:
            data = list(struct.unpack(data_type * nb_values, stream.read(nbytes)))
        return [
            data[idx * nb_components : (idx + 1) * nb_components]
            for idx in range(nb_values // nb_components)
        ]

    return None


# -----------------------------------------------------------------------------


def linspace(start, stop, num):
    delta = (stop - start) / (num - 1)
    return [start + i * delta for i in range(num)]


data_array_sha_mapping = {}


def digest(array):
    obj_id = reference_id(array)

    record = None
    if obj_id in data_array_sha_mapping:
        record = data_array_sha_mapping[obj_id]

    if record and record["mtime"] == array.GetMTime():
        return record["sha"]

    record = {"sha": hash_data_array(array), "mtime": array.GetMTime()}

    data_array_sha_mapping[obj_id] = record
    return record["sha"]


# -----------------------------------------------------------------------------


def get_range_info(array, component):
    r = array.GetRange(component)
    comp_range = {}
    comp_range["min"] = r[0]
    comp_range["max"] = r[1]
    comp_range["component"] = array.GetComponentName(component)
    return comp_range


# -----------------------------------------------------------------------------


def get_array_description(array, context, **kwargs):
    if not array:
        return None

    p_md5 = digest(array)
    context.cache_data_array(
        p_md5, {"array": array, "mTime": array.GetMTime(), "ts": time.time(), **kwargs}
    )

    root = {}
    root["hash"] = p_md5
    root["vtkClass"] = "vtkDataArray"
    root["name"] = array.GetName()
    root["dataType"] = kwargs.get("dataType", get_js_array_type(array))
    root["numberOfComponents"] = array.GetNumberOfComponents()
    root["size"] = array.GetNumberOfComponents() * array.GetNumberOfTuples()
    root["ranges"] = []
    if root["numberOfComponents"] > 1:
        for i in range(root["numberOfComponents"]):
            root["ranges"].append(get_range_info(array, i))
        root["ranges"].append(get_range_info(array, -1))
    else:
        root["ranges"].append(get_range_info(array, 0))

    return root


# -----------------------------------------------------------------------------


def extract_required_fields(
    extracted_fields, parent, dataset, context, requested_fields=["Normals", "TCoords"]
):
    arrays_to_export = set()
    export_all = "*" in requested_fields
    # Identify arrays to export
    if not export_all:
        # FIXME should evolve and support funky mapper which leverage many arrays
        if parent and parent.IsA("vtkMapper"):
            mapper = parent
            scalar_visibility = mapper.GetScalarVisibility()
            array_access_mode = mapper.GetArrayAccessMode()
            color_array_name = (
                mapper.GetArrayName() if array_access_mode == 1 else mapper.GetArrayId()
            )
            # color_mode = mapper.GetColorMode()
            scalar_mode = mapper.GetScalarMode()
            if scalar_visibility and scalar_mode in (1, 3):
                array_to_export = dataset.GetPointData().GetArray(color_array_name)
                if array_to_export is None:
                    array_to_export = dataset.GetPointData().GetScalars()
                arrays_to_export.add(array_to_export)
            if scalar_visibility and scalar_mode in (2, 4):
                array_to_export = dataset.GetCellData().GetArray(color_array_name)
                if array_to_export is None:
                    array_to_export = dataset.GetCellData().GetScalars()
                arrays_to_export.add(array_to_export)
            if scalar_visibility and scalar_mode == 0:
                array_to_export = dataset.GetPointData().GetScalars()
                if array_to_export is None:
                    array_to_export = dataset.GetCellData().GetScalars()
                arrays_to_export.add(array_to_export)

        if parent and parent.IsA("vtkTexture") and dataset.GetPointData().GetScalars():
            arrays_to_export.add(dataset.GetPointData().GetScalars())

        arrays_to_export.update(
            [
                getattr(dataset.GetPointData(), "Get" + requested_field, lambda: None)()
                for requested_field in requested_fields
            ]
        )

    # Browse all arrays
    for location, field_data in [
        ("pointData", dataset.GetPointData()),
        ("cellData", dataset.GetCellData()),
    ]:
        for array_index in range(field_data.GetNumberOfArrays()):
            array = field_data.GetArray(array_index)
            if export_all or array in arrays_to_export:
                array_meta = get_array_description(array, context)
                if array_meta:
                    array_meta["location"] = location
                    attribute = field_data.IsArrayAnAttribute(array_index)
                    array_meta["registration"] = (
                        "set" + field_data.GetAttributeTypeAsString(attribute)
                        if attribute >= 0
                        else "addArray"
                    )
                    extracted_fields.append(array_meta)



from .local_rendering import vtkWebLocalRendering
from .mouse_handler import vtkWebMouseHandler
from .publish_image_delivery import vtkWebPublishImageDelivery
from .view_port import vtkWebViewPort

__all__ = [
    "vtkWebLocalRendering",
    "vtkWebMouseHandler",
    "vtkWebPublishImageDelivery",
    "vtkWebViewPort",
]



import logging
import os
import importlib
import sys

vtk_module_name = os.environ.get("VTK_MODULE_NAME", "vtkmodules")
sys.modules["vtk_module"] = importlib.import_module(vtk_module_name)
from vtk_module.vtkFiltersGeometry import vtkDataSetSurfaceFilter

from .registry import class_name
from .serialize import serialize
from .utils import reference_id, wrap_id
from .cache import cache_properties

logger = logging.getLogger(__name__)


def generic_mapper_serializer(parent, mapper, mapper_id, context, depth):
    # This kind of mapper requires us to get 2 items: input data and lookup
    # table
    data_object = None
    data_object_instance = None
    lookup_table_instance = None
    calls = []
    dependencies = []

    if hasattr(mapper, "GetInputDataObject"):
        mapper.GetInputAlgorithm().Update()
        data_object = mapper.GetInputDataObject(0, 0)
    else:
        if context.debug_all:
            print("This mapper does not have GetInputDataObject method")

    if data_object:
        if data_object.IsA("vtkDataSet"):
            alg = vtkDataSetSurfaceFilter()
            alg.SetInputData(data_object)
            alg.Update()
            data_object = alg.GetOutput()

        data_object_id = "%s-dataset" % mapper_id
        data_object_instance = serialize(
            mapper, data_object, data_object_id, context, depth + 1
        )

        if data_object_instance:
            dependencies.append(data_object_instance)
            calls.append(["setInputData", [wrap_id(data_object_id)]])

    lookup_table = None

    if hasattr(mapper, "GetLookupTable"):
        lookup_table = mapper.GetLookupTable()
    else:
        if context.debug_all:
            print("This mapper does not have GetLookupTable method")

    if lookup_table:
        lookup_table_id = reference_id(lookup_table)
        lookup_table_instance = serialize(
            mapper, lookup_table, lookup_table_id, context, depth + 1
        )
        if lookup_table_instance:
            dependencies.append(lookup_table_instance)
            calls.append(["setLookupTable", [wrap_id(lookup_table_id)]])

    if data_object_instance:
        color_array_name = (
            mapper.GetArrayName()
            if mapper.GetArrayAccessMode() == 1
            else mapper.GetArrayId()
        )
        return {
            "parent": reference_id(parent),
            "id": mapper_id,
            "type": class_name(mapper),
            "properties": cache_properties(
                mapper_id,
                context,
                {
                    "resolveCoincidentTopology": mapper.GetResolveCoincidentTopology(),
                    "renderTime": mapper.GetRenderTime(),
                    "arrayAccessMode": mapper.GetArrayAccessMode(),
                    "scalarRange": mapper.GetScalarRange(),
                    "useLookupTableScalarRange": 1
                    if mapper.GetUseLookupTableScalarRange()
                    else 0,
                    "scalarVisibility": mapper.GetScalarVisibility(),
                    "colorByArrayName": color_array_name,
                    "colorMode": mapper.GetColorMode(),
                    "scalarMode": mapper.GetScalarMode(),
                    "interpolateScalarsBeforeMapping": 1
                    if mapper.GetInterpolateScalarsBeforeMapping()
                    else 0,
                },
            ),
            "calls": calls,
            "dependencies": dependencies,
        }

    return None


# -----------------------------------------------------------------------------


def generic_volume_mapper_serializer(parent, mapper, mapper_id, context, depth):
    # This kind of mapper requires us to get 2 items: input data and lookup
    # table
    data_object = None
    data_object_instance = None
    # lookup_table_instance = None
    calls = []
    dependencies = []

    if hasattr(mapper, "GetInputDataObject"):
        mapper.GetInputAlgorithm().Update()
        data_object = mapper.GetInputDataObject(0, 0)
    else:
        logger.debug("This mapper does not have GetInputDataObject method")

    if data_object:
        data_object_id = "%s-dataset" % mapper_id
        data_object_instance = serialize(
            mapper, data_object, data_object_id, context, depth + 1
        )

        if data_object_instance:
            dependencies.append(data_object_instance)
            calls.append(["setInputData", [wrap_id(data_object_id)]])

    if data_object_instance:
        if hasattr(mapper, "GetImageSampleDistance"):
            image_sample_distance = mapper.GetImageSampleDistance()
        else:
            image_sample_distance = 1.0
        return {
            "parent": reference_id(parent),
            "id": mapper_id,
            "type": class_name(mapper),
            "properties": cache_properties(
                mapper_id,
                context,
                {
                    # VolumeMapper
                    "sampleDistance": abs(mapper.GetSampleDistance()),
                    "imageSampleDistance": image_sample_distance,
                    # "maximumSamplesPerRay": mapper.GetMaximumSamplesPerRay(),
                    "autoAdjustSampleDistances": mapper.GetAutoAdjustSampleDistances(),
                    "blendMode": mapper.GetBlendMode(),
                    # "ipScalarRange": mapper.GetIpScalarRange(),
                    # "filterMode": mapper.GetFilterMode(),
                    # "preferSizeOverAccuracy": mapper.Get(),
                },
            ),
            "calls": calls,
            "dependencies": dependencies,
        }

    return None



def handle_property(hash_list, prop_value):
    if isinstance(prop_value, list):
        for item in prop_value:
            handle_property(hash_list, item)
        return

    if not isinstance(prop_value, dict):
        return

    if "vtkClass" in prop_value and "hash" in prop_value:
        hash_list.append(
            dict(
                hash=prop_value.get("hash"),
                type=prop_value.get("dataType"),
                vtk=prop_value.get("vtkClass"),
            )
        )


def handle_instance(hash_list, instance):
    if "dependencies" in instance:
        for child in instance.get("dependencies"):
            handle_instance(hash_list, child)
    if "properties" in instance:
        for prop in instance.get("properties").values():
            handle_property(hash_list, prop)
    return hash_list


def extract_array_hash(scene_description):
    return handle_instance([], scene_description)



from paraview import simple
from wslink import register as export_rpc

from trame_vtk.modules.vtk.serializers import (
    reference_id,
    initialize_serializers,
    serialize,
    serialize_widget,
    extract_array_hash,
    SynchronizationContext,
)

from .web_protocol import ParaViewWebProtocol


class ParaViewWebLocalRendering(ParaViewWebProtocol):
    def __init__(self, **kwargs):
        super().__init__()
        initialize_serializers()
        self.context = SynchronizationContext()
        self.tracking_views = {}
        self.mtime = 0

    # RpcName: get_array => viewport.geometry.array.get
    @export_rpc("viewport.geometry.array.get")
    def get_array(self, data_hash, binary=False):
        if binary:
            return self.addAttachment(
                self.context.get_cached_data_array(data_hash, binary)
            )
        return self.context.get_cached_data_array(data_hash, binary)

    # RpcName: add_view_observer => viewport.geometry.view.observer.add
    @export_rpc("viewport.geometry.view.observer.add")
    def add_view_observer(self, view_id):
        s_view = self.get_view(view_id)
        if not s_view:
            return {"error": "Unable to get view with id %s" % view_id}

        real_view_id = s_view.GetGlobalIDAsString()

        def push_geometry(new_subscription=False):
            simple.Render(s_view)
            state_to_return = self.get_view_state(real_view_id, new_subscription)
            state_to_return["mtime"] = 0 if new_subscription else self.mtime
            self.mtime += 1
            return state_to_return

        if real_view_id not in self.tracking_views:

            def observer_callback(*_, **__):
                return self.publish(
                    "viewport.geometry.view.subscription",
                    push_geometry(),
                )

            tag = self.app.AddObserver("UpdateEvent", observer_callback)
            self.tracking_views[real_view_id] = {"tags": [tag], "observerCount": 1}
        else:
            # There is an observer on this view already
            self.tracking_views[real_view_id]["observerCount"] += 1

        self.publish("viewport.geometry.view.subscription", push_geometry(True))
        return {"success": True, "viewId": real_view_id}

    # RpcName: remove_view_observer => viewport.geometry.view.observer.remove
    @export_rpc("viewport.geometry.view.observer.remove")
    def remove_view_observer(self, view_id):
        s_view = self.get_view(view_id)
        if not s_view:
            return {"error": "Unable to get view with id %s" % view_id}

        real_view_id = s_view.GetGlobalIDAsString()

        observer_info = None
        if real_view_id in self.tracking_views:
            observer_info = self.tracking_views[real_view_id]

        if not observer_info:
            return {"error": "Unable to find subscription for view %s" % real_view_id}

        observer_info["observerCount"] -= 1

        if observer_info["observerCount"] <= 0:
            for tag in observer_info["tags"]:
                self.app.RemoveObserver(tag)
            del self.tracking_views[real_view_id]

        return {"result": "success"}

    # RpcName: get_view_state => viewport.geometry.view.get.state
    @export_rpc("viewport.geometry.view.get.state")
    def get_view_state(
        self,
        view_id,
        new_subscription=False,
        widgets=None,
        orientation_axis=0,
        **kwargs,
    ):
        s_view = self.get_view(view_id)
        if not s_view:
            return {"error": "Unable to get view with id %s" % view_id}

        self.context.set_ignore_last_dependencies(new_subscription)

        # Get the active view and render window, use it to iterate over renderers
        render_window = s_view.GetRenderWindow()
        render_window_id = s_view.GetGlobalIDAsString()
        view_instance = serialize(
            None, render_window, render_window_id, self.context, 1
        )
        view_instance["extra"] = {
            "vtkRefId": reference_id(render_window),
            "centerOfRotation": s_view.CenterOfRotation.GetData(),
            "camera": reference_id(s_view.GetActiveCamera()),
        }

        # Handle widgets/behaviors
        if widgets:
            behaviors = {}
            view_instance["behaviors"] = behaviors
            for widget in widgets:
                serialize_widget(behaviors, widget)
        elif orientation_axis:
            view_instance["behaviors"] = {
                "autoOrientation": 1,
            }

        self.context.set_ignore_last_dependencies(False)
        self.context.check_for_arrays_to_release()

        if view_instance:
            return view_instance

        return None

    @export_rpc("viewport.geometry.view.get.export")
    def get_standalone_state(self, view_id, widgets=None, orientation_axis=0, **kwargs):
        scene_description = self.get_view_state(
            view_id,
            new_subscription=True,
            widgets=widgets,
            orientation_axis=orientation_axis,
            **kwargs,
        )
        hashes = {}
        for entry in extract_array_hash(scene_description):
            data_hash = entry.get("hash")
            hashes[data_hash] = dict(
                **entry, content=self.context.get_cached_data_array(data_hash, False)
            )

        return dict(hashes=hashes, scene=scene_description)



from paraview import servermanager


def apply_default_interaction_settings():
    # ProxyManager helper
    pxm = servermanager.ProxyManager()

    # Update interaction mode
    interaction_proxy = pxm.GetProxy("settings", "RenderViewInteractionSettings")
    interaction_proxy.Camera3DManipulators = [
        "Rotate",
        "Pan",
        "Zoom",  # -
        "Pan",
        "Roll",
        "Pan",  # shift
        "Zoom",
        "Rotate",
        "Zoom",  # ctrl
    ]

    # Custom rendering settings
    rendering_settings = pxm.GetProxy("settings", "RenderViewSettings")
    rendering_settings.LODThreshold = 102400



from typing import Type, Callable, Dict, Optional
from functools import partialmethod
import os
import importlib
import sys

vtk_module_name = os.environ.get("VTK_MODULE_NAME", "vtkmodules")
sys.modules["vtk_module"] = importlib.import_module(vtk_module_name)

from vtk_module.vtkInteractionWidgets import (
    vtkAbstractWidget,
    vtkWidgetRepresentation,
)

from vtk_module.vtkCommonCore import vtkCommand

EventCallback = Callable[[vtkCommand.EventIds, Optional["VtkWidget"]], None]


def callback_wrapper(target: "VtkWidget", callback: EventCallback):
    def wrapped_callback(*args, **kwargs):
        callback(vtkCommand.GetEventIdFromString(args[1]), target, *args[2:], **kwargs)

    return wrapped_callback


class VtkWidget:
    def __init__(self, w: vtkAbstractWidget):
        self._w = w
        self._listeners: Dict[vtkCommand.EventIds, Dict[EventCallback, int]] = {}

    @property
    def vtk_widget(self) -> vtkAbstractWidget:
        return self._w

    @property
    def vtk_representation(self) -> vtkWidgetRepresentation:
        return self._w.GetRepresentation()

    def __getattr__(self, name: str):
        if hasattr(self.vtk_widget, name):
            return getattr(self.vtk_widget, name)
        elif hasattr(self.vtk_representation, name):
            return getattr(self.vtk_representation, name)
        else:
            return super().__getattribute__(name)

    def enable(self):
        self._w.On()

    def disable(self):
        self._w.Off()

    def add_event_listener(
        self, event: vtkCommand.EventIds, listener: EventCallback
    ) -> Callable[[], None]:
        listeners = self._listeners.setdefault(event, {})

        tag = self.vtk_widget.AddObserver(event, callback_wrapper(self, listener))

        listeners[listener] = tag

        def listener_remover():
            self.remove_event_listener(event, listener)

        return listener_remover

    def remove_event_listener(
        self, event: vtkCommand.EventIds, listener: Optional[EventCallback] = None
    ):
        listeners = self._listeners.get(event, {})

        if listener is None:
            for tag in listeners.values():
                self.vtk_widget.RemoveObserver(tag)

            listeners.clear()
        else:
            tag = listeners.pop(listener, None)

            if tag is not None:
                self.vtk_widget.RemoveObserver(tag)


def _to_snake_case(s: str) -> str:
    out = ""

    for c in s:
        if c.isupper():
            out += f"_{c.lower()}"
        else:
            out += c

    return out


# Dynamically add methods to register listeners for each event available on vtkCommand
# For example:
#   - `vtkCommant.InteractionEvent` becomes `VtkWidget.on_interaction()`
#   - `vtkCommant.StartInteractionEvent` becomes `VtkWidget.on_start_interaction()`
#   - etc.
for name in vars(vtkCommand):
    value = getattr(vtkCommand, name)

    if isinstance(value, vtkCommand.EventIds):
        setattr(
            VtkWidget,
            _to_snake_case("on" + name.replace("Event", "")),
            partialmethod(VtkWidget.add_event_listener, value),
        )


WidgetParam = vtkAbstractWidget | Type[vtkAbstractWidget]
RepresentationParam = vtkWidgetRepresentation | Type[vtkWidgetRepresentation]


class WidgetManager:
    def __init__(self, renderer):
        self._renderer = renderer

    @property
    def _window(self):
        return self._renderer.GetRenderWindow()

    @property
    def _interactor(self):
        return self._window.GetInteractor()

    def add_widget(
        self, w: WidgetParam, r: Optional[RepresentationParam] = None
    ) -> VtkWidget:
        raw_widget: vtkAbstractWidget
        raw_representation: Optional[vtkWidgetRepresentation] = None

        if isinstance(r, type) and issubclass(r, vtkWidgetRepresentation):
            raw_representation = r()
        elif isinstance(r, vtkWidgetRepresentation):
            raw_representation = r
        elif r is not None:
            raise TypeError(
                "The r parameter should be one of: vtkWidgetRepresentation | Type[vtkWidgetRepresentation]"
            )

        if isinstance(w, type) and issubclass(w, vtkAbstractWidget):
            raw_widget = w()
            raw_widget.CreateDefaultRepresentation()
        elif isinstance(w, vtkAbstractWidget):
            raw_widget = w
        else:
            raise TypeError(
                "The w parameter should be one of: vtkAbstractWidget | Type[vtkAbstractWidget]"
            )

        if raw_representation is not None:
            raw_widget.SetRepresentation(raw_representation)

        raw_widget.SetInteractor(self._interactor)

        widget = VtkWidget(raw_widget)

        return widget



from trame_client.utils.version import get_version

__version__ = get_version("trame-vtk")


def reference_id(ref):
    if ref:
        try:
            return ref.__this__[1:17]
        except Exception:
            id_str = str(ref)[-12:-1]
            # print('====> fallback ID %s for %s' % (id_str, ref))
            return id_str
    return "0x0"



from .registry import class_name
from .serialize import serialize
from .utils import reference_id, wrap_id
from .cache import cache_properties


def property_serializer(parent, prop_obj, prop_obj_id, context, depth):
    representation = (
        prop_obj.GetRepresentation() if hasattr(prop_obj, "GetRepresentation") else 2
    )
    color_to_use = (
        prop_obj.GetDiffuseColor()
        if hasattr(prop_obj, "GetDiffuseColor")
        else [1, 1, 1]
    )
    if representation == 1 and hasattr(prop_obj, "GetColor"):
        color_to_use = prop_obj.GetColor()

    return {
        "parent": reference_id(parent),
        "id": prop_obj_id,
        "type": class_name(prop_obj),
        "properties": cache_properties(
            prop_obj_id,
            context,
            {
                "representation": representation,
                "diffuseColor": color_to_use,
                "color": prop_obj.GetColor(),
                "ambientColor": prop_obj.GetAmbientColor(),
                "specularColor": prop_obj.GetSpecularColor(),
                "edgeColor": prop_obj.GetEdgeColor(),
                "ambient": prop_obj.GetAmbient(),
                "diffuse": prop_obj.GetDiffuse(),
                "specular": prop_obj.GetSpecular(),
                "specularPower": prop_obj.GetSpecularPower(),
                "opacity": prop_obj.GetOpacity(),
                "interpolation": prop_obj.GetInterpolation(),
                "edgeVisibility": 1 if prop_obj.GetEdgeVisibility() else 0,
                "backfaceCulling": 1 if prop_obj.GetBackfaceCulling() else 0,
                "frontfaceCulling": 1 if prop_obj.GetFrontfaceCulling() else 0,
                "pointSize": prop_obj.GetPointSize(),
                "lineWidth": prop_obj.GetLineWidth(),
                "lighting": 1 if prop_obj.GetLighting() else 0,
            },
        ),
    }


def volume_property_serializer(parent, prop_obj, prop_obj_id, context, depth):
    calls = []
    dependencies = []

    # Color handling
    lut = prop_obj.GetRGBTransferFunction()
    if lut:
        lookup_table_id = reference_id(lut)
        lookup_table_instance = serialize(
            prop_obj, lut, lookup_table_id, context, depth + 1
        )

        if lookup_table_instance:
            dependencies.append(lookup_table_instance)
            calls.append(["setRGBTransferFunction", [0, wrap_id(lookup_table_id)]])

    # Piecewise handling
    pwf = prop_obj.GetScalarOpacity()
    if pwf:
        pwf_id = reference_id(pwf)
        pwf_instance = serialize(prop_obj, pwf, pwf_id, context, depth + 1)

        if pwf_instance:
            dependencies.append(pwf_instance)
            calls.append(["setScalarOpacity", [0, wrap_id(pwf_id)]])

    return {
        "parent": reference_id(parent),
        "id": prop_obj_id,
        "type": class_name(prop_obj),
        "properties": {
            "independentComponents": prop_obj.GetIndependentComponents(),
            "interpolationType": prop_obj.GetInterpolationType(),
            "shade": prop_obj.GetShade(),
            "ambient": prop_obj.GetAmbient(),
            "diffuse": prop_obj.GetDiffuse(),
            "specular": prop_obj.GetSpecular(),
            "specularPower": prop_obj.GetSpecularPower(),
            # "useLabelOutline": prop_obj.GetUseLabelOutline(),
            # "labelOutlineThickness": prop_obj.GetLabelOutlineThickness(),
        },
        "calls": calls,
        "dependencies": dependencies,
    }



import logging

from .registry import class_name, SERIALIZERS
from .widgets import handle_widget

__all__ = ["serialize", "serialize_widget"]

logger = logging.getLogger(__name__)

# Keep track of which warnings have been printed
NO_SERIALIZER_FOR_INSTANCE = {}


def serialize(parent, instance, instance_id, context, depth):
    instance_type = class_name(instance)
    serializer = SERIALIZERS[instance_type] if instance_type in SERIALIZERS else None

    if serializer:
        return serializer(parent, instance, instance_id, context, depth)

    if instance_type not in NO_SERIALIZER_FOR_INSTANCE:
        # Only print the warning once for each type of serializer
        logger.warning(f"!!!No serializer for {instance_type} with id {instance_id}")
        NO_SERIALIZER_FOR_INSTANCE[instance_type] = instance_id

    return None


def serialize_widget(dict_out, widget):
    handle_widget(dict_out, widget)



import os

from paraview import simple
from paraview.servermanager import vtkSMTransferFunctionManager

from trame_vtk.modules.vtk.protocols.web_protocol import vtkWebProtocol


class ParaViewWebProtocol(vtkWebProtocol):
    def __init__(self):
        self.core_server = None
        self.multi_root = False
        self.base_directory = ""
        self.base_directory_map = {}

    def map_id_to_proxy(self, id):
        """
        Maps global-id for a proxy to the proxy instance. May return None if the
        id is not valid.
        """
        try:
            id = int(id)
        except ValueError:
            return None
        if id <= 0:
            return None
        return simple.servermanager._getPyProxy(
            simple.servermanager.ActiveConnection.Session.GetRemoteObject(id)
        )

    def get_view(self, vid):
        """
        Returns the view for a given view ID, if vid is None then return the
        current active view.
        :param vid: The view ID
        :type vid: str
        """
        view = self.map_id_to_proxy(vid)
        if not view:
            # Use active view is none provided.
            view = simple.GetActiveView()

        if not view:
            raise Exception(f"no view provided: {vid}")

        return view

    def debug(self, msg):
        if self.debug_mode:
            print(msg)

    def set_base_directory(self, base_path):
        self.override_data_dir_key = None
        self.base_directory = ""
        self.base_directory_map = {}
        self.multi_root = False

        if base_path.find("|") < 0:
            if base_path.find("=") >= 0:
                base_pair = base_path.split("=")
                if os.path.exists(base_pair[1]):
                    self.base_directory = base_pair[1]
                    self.override_data_dir_key = base_pair[0]
            else:
                self.base_directory = base_path
            self.base_directory = os.path.normpath(self.base_directory)
        else:
            base_dirs = base_path.split("|")
            for base_dir in base_dirs:
                base_pair = base_dir.split("=")
                if os.path.exists(base_pair[1]):
                    self.base_directory_map[base_pair[0]] = os.path.normpath(
                        base_pair[1]
                    )

            # Check if we ended up with just a single directory
            bd_keys = list(self.base_directory_map)
            if len(bd_keys) == 1:
                self.base_directory = os.path.normpath(
                    self.base_directory_map[bd_keys[0]]
                )
                self.override_data_dir_key = bd_keys[0]
                self.base_directory_map = {}
            elif len(bd_keys) > 1:
                self.multi_root = True

    def get_absolute_path(self, relative_path):
        absolute_path = None

        if self.multi_root:
            rel_path_parts = relative_path.replace("\\", "/").split("/")
            real_base_path = self.base_directory_map[rel_path_parts[0]]
            absolute_path = os.path.join(real_base_path, *rel_path_parts[1:])
        else:
            absolute_path = os.path.join(self.base_directory, relative_path)

        cleaned_path = os.path.normpath(absolute_path)

        # Make sure the cleaned_path is part of the allowed ones
        if self.multi_root:
            for key, value in self.base_directory_map.items():
                if cleaned_path.startswith(value):
                    return cleaned_path
        elif cleaned_path.startswith(self.base_directory):
            return cleaned_path

        return None

    def update_scalar_bars(self, view=None, mode=1):
        """
        Manage scalarbar state

            view:
                A view proxy or the current active view will be used.

            mode:
                HIDE_UNUSED_SCALAR_BARS = 0x01,
                SHOW_USED_SCALAR_BARS = 0x02
        """
        v = view or self.get_view(-1)
        lut_mgr = vtkSMTransferFunctionManager()
        lut_mgr.UpdateScalarBars(v.SMProxy, mode)

    def publish(self, topic, event):
        if self.core_server:
            self.core_server.publish(topic, event)



import base64
import math
import time

from paraview import simple
from wslink import register as export_rpc
from wslink import schedule_callback

from vtkmodules.vtkWebCore import vtkWebInteractionEvent

from .web_protocol import ParaViewWebProtocol


def apply_modifiers(event, interactor):
    interactor.SetShiftKey(1 if event.get("shiftKey") else 0)
    interactor.SetControlKey(1 if event.get("ctrlKey") else 0)
    interactor.SetAltKey(1 if event.get("altKey") else 0)
    # interactor.SetMetaKey(1 if event.get("metaKey") else 0)


class ParaViewWebPublishImageDelivery(ParaViewWebProtocol):
    def __init__(self, decode=True, **kwargs):
        super().__init__()
        self.tracking_views = {}
        self.last_stale_time = {}
        self.stale_handler_count = {}
        self.delta_stale_time_before_render = 0.1  # 0.1s
        self.stale_count_limit = 10
        self.decode = decode
        self.views_in_animations = []
        self.target_frame_rate = 30.0
        self.min_frame_rate = 12.0
        self.max_frame_rate = 30.0

        # Camera link handling
        self.linked_views = []
        self.link_names = []
        self.on_link_change = None

        # Mouse handling
        self.last_action = "up"
        self.active_view_id = None

    # In case some external protocol wants to monitor when link views change
    def set_link_change_callback(self, fn):
        self.on_link_change = fn

    def push_render(self, v_id, ignore_animation=False, stale_count=0):
        if v_id not in self.tracking_views:
            return

        if not self.tracking_views[v_id]["enabled"]:
            return

        if not ignore_animation and len(self.views_in_animations) > 0:
            return

        if "originalSize" not in self.tracking_views[v_id]:
            view = self.get_view(v_id)
            self.tracking_views[v_id]["originalSize"] = (
                int(view.ViewSize[0] + 0.5),
                int(view.ViewSize[1] + 0.5),
            )

        if "ratio" not in self.tracking_views[v_id]:
            self.tracking_views[v_id]["ratio"] = 1

        ratio = self.tracking_views[v_id]["ratio"]
        mtime = self.tracking_views[v_id]["mtime"]
        quality = self.tracking_views[v_id]["quality"]
        size = [
            int((s * ratio) + 0.5) for s in self.tracking_views[v_id]["originalSize"]
        ]

        reply = self.still_render(
            {"view": v_id, "mtime": mtime, "quality": quality, "size": size}
        )

        # View might have been deleted
        if not reply:
            return

        stale = reply["stale"]
        if reply["image"]:
            # depending on whether the app has encoding enabled:
            if self.decode:
                reply["image"] = base64.standard_b64decode(reply["image"])

            reply["image"] = self.addAttachment(reply["image"])
            reply["format"] = "jpeg"
            # save mtime for next call.
            self.tracking_views[v_id]["mtime"] = reply["mtime"]
            # echo back real ID, instead of -1 for 'active'
            reply["id"] = v_id
            self.publish("viewport.image.push.subscription", reply)
        if stale:
            self.last_stale_time[v_id] = time.time()
            if self.stale_handler_count[v_id] == 0:
                self.stale_handler_count[v_id] += 1
                schedule_callback(
                    self.delta_stale_time_before_render,
                    lambda: self.render_stale_image(v_id, stale_count),
                )
        else:
            self.last_stale_time[v_id] = 0

    def render_stale_image(self, v_id, stale_count=0):
        if v_id in self.stale_handler_count and self.stale_handler_count[v_id] > 0:
            self.stale_handler_count[v_id] -= 1

            if self.last_stale_time[v_id] != 0:
                delta = time.time() - self.last_stale_time[v_id]
                # Break on stale_count otherwise linked view will always report to be stale
                # And loop forever
                if (
                    delta >= (self.delta_stale_time_before_render * (stale_count + 1))
                    and stale_count < self.stale_count_limit
                ):
                    self.push_render(v_id, False, stale_count + 1)
                elif delta < self.delta_stale_time_before_render:
                    self.stale_handler_count[v_id] += 1
                    schedule_callback(
                        self.delta_stale_time_before_render - delta + 0.001,
                        lambda: self.render_stale_image(v_id, stale_count),
                    )

    def animate(self, render_all_views=True):
        if len(self.views_in_animations) == 0:
            return

        next_animate_time = time.time() + 1.0 / self.target_frame_rate

        # Handle the rendering of the views
        if self.active_view_id:
            self.push_render(self.active_view_id, True)

        if render_all_views:
            for v_id in set(self.views_in_animations):
                if v_id != self.active_view_id:
                    self.push_render(v_id, True)

        next_animate_time -= time.time()

        if self.target_frame_rate > self.max_frame_rate:
            self.target_frame_rate = self.max_frame_rate

        if next_animate_time < 0:
            if next_animate_time < -1.0:
                self.target_frame_rate = 1
            if self.target_frame_rate > self.min_frame_rate:
                self.target_frame_rate -= 1.0
            if self.active_view_id:
                # If active view, prioritize that one over the others
                # -> Divide by 2 the refresh rate of the other views
                schedule_callback(0.001, lambda: self.animate(not render_all_views))
            else:
                # Keep animating at the best rate we can
                schedule_callback(0.001, lambda: self.animate())
        else:
            # We have time so let's render all
            if (
                self.target_frame_rate < self.max_frame_rate
                and next_animate_time > 0.005
            ):
                self.target_frame_rate += 1.0
            schedule_callback(next_animate_time, lambda: self.animate())

    @export_rpc("viewport.image.animation.fps.max")
    def set_max_frame_rate(self, fps=30):
        self.max_frame_rate = fps

    @export_rpc("viewport.image.animation.fps.get")
    def get_current_frame_rate(self):
        return self.target_frame_rate

    @export_rpc("viewport.image.animation.start")
    def start_view_animation(self, view_id="-1"):
        s_view = self.get_view(view_id)
        real_view_id = s_view.GetGlobalIDAsString()

        self.views_in_animations.append(real_view_id)
        if len(self.views_in_animations) == 1:
            self.animate()

    @export_rpc("viewport.image.animation.stop")
    def stop_view_animation(self, view_id="-1"):
        s_view = self.get_view(view_id)
        real_view_id = s_view.GetGlobalIDAsString()

        if (
            real_view_id in self.views_in_animations
            and real_view_id in self.tracking_views
        ):
            progress_rendering = self.tracking_views[real_view_id]["streaming"]
            self.views_in_animations.remove(real_view_id)
            if progress_rendering:
                self.progressive_render(real_view_id)

    def progressive_render(self, view_id="-1"):
        s_view = self.get_view(view_id)
        real_view_id = s_view.GetGlobalIDAsString()

        if real_view_id in self.views_in_animations:
            return

        if s_view.GetSession().GetPendingProgress():
            schedule_callback(
                self.delta_stale_time_before_render,
                lambda: self.progressive_render(view_id),
            )
        else:
            again = s_view.StreamingUpdate(True)
            self.push_render(real_view_id, True)

            if again:
                schedule_callback(0.001, lambda: self.progressive_render(view_id))

    @export_rpc("viewport.image.push")
    def image_push(self, options):
        view = self.get_view(options["view"])
        view_id = view.GetGlobalIDAsString()

        # Make sure an image is pushed
        self.app.InvalidateCache(view.SMProxy)

        self.push_render(view_id)

    # Internal function since the reply[image] is not
    # JSON(serializable) it can not be an RPC one
    def still_render(self, options):
        """
        RPC Callback to render a view and obtain the rendered image.
        """
        begin_time = int(round(time.time() * 1000))
        view_id = str(options["view"])
        view = self.get_view(view_id)

        # If no view id provided, skip rendering
        if not view_id:
            print("No view")
            print(options)
            return None

        # Make sure request match our selected view
        if view_id != "-1" and view.GetGlobalIDAsString() != view_id:
            # We got active view rather than our request
            view = None

        # No view to render => need some cleanup
        if not view:
            # The view has been deleted, we can not render it...
            # Clean up old view state
            if view_id in self.views_in_animations:
                self.views_in_animations.remove(view_id)

            if view_id in self.tracking_views:
                del self.tracking_views[view_id]

            if view_id in self.stale_handler_count:
                del self.stale_handler_count[view_id]

            # the view does not exist anymore, skip rendering
            return None

        # We are in business to render our view...

        # Make sure our view size match our request
        size = view.ViewSize[0:2]
        resize = size != options.get("size", size)
        if resize:
            size = options["size"]
            if size[0] > 10 and size[1] > 10:
                view.ViewSize = size

        # Rendering options
        t = 0
        if options and "mtime" in options:
            t = options["mtime"]
        quality = 100
        if options and "quality" in options:
            quality = options["quality"]
        local_time = 0
        if options and "localTime" in options:
            local_time = options["localTime"]
        reply = {}
        app = self.app
        if t == 0:
            app.InvalidateCache(view.SMProxy)
        if self.decode:
            still_render = app.StillRenderToString
        else:
            still_render = app.StillRenderToBuffer
        reply_image = still_render(view.SMProxy, t, quality)

        # Check that we are getting image size we have set if not wait until we
        # do. The render call will set the actual window size.
        tries = 10
        while (
            resize
            and list(app.GetLastStillRenderImageSize()) != size
            and size != [0, 0]
            and tries > 0
        ):
            app.InvalidateCache(view.SMProxy)
            reply_image = still_render(view.SMProxy, t, quality)
            tries -= 1

        if (
            not resize
            and options
            and ("clearCache" in options)
            and options["clearCache"]
        ):
            app.InvalidateCache(view.SMProxy)
            reply_image = still_render(view.SMProxy, t, quality)

        # Pack the result
        reply["stale"] = app.GetHasImagesBeingProcessed(view.SMProxy)
        reply["mtime"] = app.GetLastStillRenderToMTime()
        reply["size"] = view.ViewSize[0:2]
        reply["memsize"] = reply_image.GetDataSize() if reply_image else 0
        reply["format"] = "jpeg;base64" if self.decode else "jpeg"
        reply["global_id"] = view.GetGlobalIDAsString()
        reply["localTime"] = local_time
        if self.decode:
            reply["image"] = reply_image
        else:
            # Convert the vtkUnsignedCharArray into a bytes object, required by Autobahn websockets
            reply["image"] = memoryview(reply_image).tobytes() if reply_image else None

        end_time = int(round(time.time() * 1000))
        reply["workTime"] = end_time - begin_time

        return reply

    @export_rpc("viewport.image.push.observer.add")
    def add_render_observer(self, view_id):
        s_view = self.get_view(view_id)
        if not s_view:
            return {"error": "Unable to get view with id %s" % view_id}

        real_view_id = s_view.GetGlobalIDAsString()

        if real_view_id not in self.tracking_views:

            def observer_callback(*_, **__):
                return self.push_render(real_view_id)

            def start_callback(*_, **__):
                return self.start_view_animation(real_view_id)

            def stop_callback(*_, **__):
                return self.stop_view_animation(real_view_id)

            tag = self.app.AddObserver("UpdateEvent", observer_callback)
            tag_start = self.app.AddObserver("StartInteractionEvent", start_callback)
            tag_stop = self.app.AddObserver("EndInteractionEvent", stop_callback)
            # TODO do we need self.app.AddObserver('ResetActiveView', reset_active_view())
            self.tracking_views[real_view_id] = {
                "tags": [tag, tag_start, tag_stop],
                "observerCount": 1,
                "mtime": 0,
                "enabled": True,
                "quality": 100,
                "streaming": s_view.GetClientSideObject().GetEnableStreaming(),
            }
            self.stale_handler_count[real_view_id] = 0
        else:
            # There is an observer on this view already
            self.tracking_views[real_view_id]["observerCount"] += 1

        self.push_render(real_view_id)
        return {"success": True, "viewId": real_view_id}

    @export_rpc("viewport.image.push.observer.remove")
    def remove_render_observer(self, view_id):
        s_view = None
        try:
            s_view = self.get_view(view_id)
        except Exception:
            print("no view with ID %s available in remove_render_observer" % view_id)

        real_view_id = s_view.GetGlobalIDAsString() if s_view else view_id

        observer_info = None
        if real_view_id in self.tracking_views:
            observer_info = self.tracking_views[real_view_id]

        if not observer_info:
            return {"error": "Unable to find subscription for view %s" % real_view_id}

        observer_info["observerCount"] -= 1

        if observer_info["observerCount"] <= 0:
            for tag in observer_info["tags"]:
                self.app.RemoveObserver(tag)
            del self.tracking_views[real_view_id]
            del self.stale_handler_count[real_view_id]

        return {"result": "success"}

    @export_rpc("viewport.image.push.quality.get")
    def get_view_quality(self, view_id):
        response = dict(quality=1, ratio=1)
        s_view = self.get_view(view_id)

        if s_view:
            real_view_id = s_view.GetGlobalIDAsString()
            if real_view_id in self.tracking_views:
                observer_info = self.tracking_views[real_view_id]
                response["quality"] = observer_info.get("quality", 100)
                response["ratio"] = observer_info.get("ratio", 1)

        return response

    @export_rpc("viewport.image.push.quality")
    def set_view_quality(self, view_id, quality, ratio=1, update_linked_view=True):
        s_view = self.get_view(view_id)
        if not s_view:
            return {"error": "Unable to get view with id %s" % view_id}

        real_view_id = s_view.GetGlobalIDAsString()
        observer_info = None
        if real_view_id in self.tracking_views:
            observer_info = self.tracking_views[real_view_id]

        if not observer_info:
            return {"error": "Unable to find subscription for view %s" % real_view_id}

        observer_info["quality"] = quality
        observer_info["ratio"] = ratio

        # Handle linked view quality/ratio synch
        if update_linked_view and real_view_id in self.linked_views:
            for vid in self.linked_views:
                self.set_view_quality(vid, quality, ratio, False)

        # Update image size right now!
        if "originalSize" in self.tracking_views[real_view_id]:
            size = [
                int((s * ratio) + 0.5)
                for s in self.tracking_views[real_view_id]["originalSize"]
            ]
            if "SetSize" in s_view:
                s_view.SetSize(size)
            else:
                s_view.ViewSize = size

        return {"result": "success"}

    @export_rpc("viewport.image.push.original.size")
    def set_view_size(self, view_id, width, height):
        if width < 10 or height < 10:
            return {"result": "size skip"}

        s_view = self.get_view(view_id)
        if not s_view:
            return {"error": "Unable to get view with id %s" % view_id}

        real_view_id = s_view.GetGlobalIDAsString()
        observer_info = None
        if real_view_id in self.tracking_views:
            observer_info = self.tracking_views[real_view_id]

        if not observer_info:
            return {"error": "Unable to find subscription for view %s" % real_view_id}

        observer_info["originalSize"] = (int(width + 0.5), int(height + 0.5))

        return {"result": "success"}

    @export_rpc("viewport.image.push.enabled")
    def enable_view(self, view_id, enabled):
        s_view = self.get_view(view_id)
        if not s_view:
            return {"error": "Unable to get view with id %s" % view_id}

        real_view_id = s_view.GetGlobalIDAsString()
        observer_info = None
        if real_view_id in self.tracking_views:
            observer_info = self.tracking_views[real_view_id]

        if not observer_info:
            return {"error": "Unable to find subscription for view %s" % real_view_id}

        observer_info["enabled"] = enabled

        return {"result": "success"}

    @export_rpc("viewport.image.push.invalidate.cache")
    def invalidate_cache(self, view_id):
        s_view = self.get_view(view_id)
        if not s_view:
            return {"error": "Unable to get view with id %s" % view_id}

        self.app.InvalidateCache(s_view.SMProxy)
        self.app.InvokeEvent("UpdateEvent")
        return {"result": "success"}

    # -------------------------------------------------------------------------
    # View linked
    # -------------------------------------------------------------------------

    def validate_view_links(self):
        for link_name in self.link_names:
            simple.RemoveCameraLink(link_name)
        self.link_names = []

        if len(self.linked_views) > 1:
            view_list = [self.get_view(vid) for vid in self.linked_views]
            ref_view = view_list.pop(0)
            for view in view_list:
                link_name = "%s_%s" % (
                    ref_view.GetGlobalIDAsString(),
                    view.GetGlobalIDAsString(),
                )
                simple.AddCameraLink(ref_view, view, link_name)
                self.link_names.append(link_name)

            # Synch camera state
            src_view = view_list[0]
            dst_views = view_list[1:]
            _push_camera_link(src_view, dst_views)

    @export_rpc("viewport.view.link")
    def update_view_link(self, view_id=None, link_state=False):
        if view_id:
            if link_state:
                self.linked_views.append(view_id)
            else:
                try:
                    self.linked_views.remove(view_id)
                except Exception:
                    pass
            # self.validate_view_links()

        if len(self.linked_views) > 1:
            all_views = [self.get_view(vid) for vid in self.linked_views]
            _push_camera_link(all_views[0], all_views[1:])

        if self.on_link_change:
            self.on_link_change(self.linked_views)

        if link_state:
            self.app.InvokeEvent("UpdateEvent")

        return self.linked_views

    # -------------------------------------------------------------------------
    # Mouse handling
    # -------------------------------------------------------------------------

    @export_rpc("viewport.mouse.interaction")
    def mouse_interaction(self, event):
        """
        RPC Callback for mouse interactions.
        """
        if "x" not in event or "y" not in event:
            return 0

        view = self.get_view(event["view"])

        if hasattr(view, "UseInteractiveRenderingForScreenshots"):
            if event["action"] == "down":
                view.UseInteractiveRenderingForScreenshots = 1
            elif event["action"] == "up":
                view.UseInteractiveRenderingForScreenshots = 0

        buttons = 0
        if event["buttonLeft"]:
            buttons |= vtkWebInteractionEvent.LEFT_BUTTON
        if event["buttonMiddle"]:
            buttons |= vtkWebInteractionEvent.MIDDLE_BUTTON
        if event["buttonRight"]:
            buttons |= vtkWebInteractionEvent.RIGHT_BUTTON

        modifiers = 0
        if event["shiftKey"]:
            modifiers |= vtkWebInteractionEvent.SHIFT_KEY
        if event["ctrlKey"]:
            modifiers |= vtkWebInteractionEvent.CTRL_KEY
        if event["altKey"]:
            modifiers |= vtkWebInteractionEvent.ALT_KEY
        if event["metaKey"]:
            modifiers |= vtkWebInteractionEvent.META_KEY

        pvevent = vtkWebInteractionEvent()
        pvevent.SetButtons(buttons)
        pvevent.SetModifiers(modifiers)
        pvevent.SetX(event["x"])
        pvevent.SetY(event["y"])
        # pvevent.SetKeyCode(event["charCode"])
        ret_val = self.app.HandleInteractionEvent(view.SMProxy, pvevent)
        del pvevent

        self.active_view_id = view.GetGlobalIDAsString()

        if event["action"] == "down" and self.last_action != event["action"]:
            self.app.InvokeEvent("StartInteractionEvent")

        if event["action"] == "up" and self.last_action != event["action"]:
            self.app.InvokeEvent("EndInteractionEvent")

        # if ret_val :
        #  self.app.InvokeEvent('UpdateEvent')

        if self.active_view_id in self.linked_views:
            dst_views = [self.get_view(vid) for vid in self.linked_views]
            _push_camera_link(view, dst_views)

        self.last_action = event["action"]

        return ret_val

    @export_rpc("viewport.mouse.zoom.wheel")
    def update_zoom_from_wheel(self, event):
        view_proxy = self.get_view(event["view"])

        if not view_proxy:
            return

        interactor = view_proxy.GetInteractor()
        if not interactor:
            # Can't do anything.
            return

        apply_modifiers(event, interactor)

        if "x" in event and "y" in event and interactor.GetRenderWindow():
            # Set the mouse position, so that if there are multiple
            # renderers, the interactor can figure out which one should
            # be modified.
            view_size = interactor.GetRenderWindow().GetSize()
            pos_x = math.floor(view_size[0] * event["x"] + 0.5)
            pos_y = math.floor(view_size[1] * event["y"] + 0.5)

            interactor.SetEventPosition(pos_x, pos_y)
            interactor.MouseMoveEvent()

        if "Start" in event["type"]:
            self.app.InvokeEvent("StartInteractionEvent")
            # It seems every time a StartInteractionEvent is sent, a
            # mouse wheel event with the same spin is sent afterward. We
            # don't want to zoom twice, so do not perform a zoom on the
            # StartInteractionEvent.
            return

        if "End" in event["type"]:
            self.app.InvokeEvent("EndInteractionEvent")
            # This is done
            return

        spin_y = event.get("spinY", 0)
        direction = "Backward" if spin_y >= 0 else "Forward"
        method = getattr(interactor, f"MouseWheel{direction}Event")

        style = interactor.GetInteractorStyle()
        prev_motion_factor = style.GetMouseWheelMotionFactor()
        style.SetMouseWheelMotionFactor(prev_motion_factor * abs(spin_y))
        try:
            method()
        finally:
            style.SetMouseWheelMotionFactor(prev_motion_factor)

        view_proxy.UpdatePropertyInformation()

        root_id = view_proxy.GetGlobalIDAsString()
        if root_id in self.linked_views:
            dst_views = [self.get_view(vid) for vid in self.linked_views]
            _push_camera_link(view_proxy, dst_views)


CAMERA_PROP_NAMES = [
    "CameraFocalPoint",
    "CameraParallelProjection",
    "CameraParallelScale",
    "CameraPosition",
    "CameraViewAngle",
    "CameraViewUp",
]


def _push_camera_link(view_src, view_dst_list):
    props = {}
    for name in CAMERA_PROP_NAMES:
        props[name] = getattr(view_src, name)
    for v in view_dst_list:
        for name in CAMERA_PROP_NAMES:
            v.__setattr__(name, props[name])
    return props



import logging

from .registry import class_name
from .serialize import serialize
from .utils import reference_id, rgb_float_to_hex, wrap_id
from .cache import cache_properties, get_cached_property

logger = logging.getLogger(__name__)


def generic_actor_serializer(parent, actor, actor_id, context, depth):
    # This kind of actor has two "children" of interest, a property and a
    # mapper
    actor_visibility = actor.GetVisibility()
    mapper_instance = None
    property_instance = None
    calls = []
    dependencies = []
    add_on = {}

    if actor_visibility:
        mapper = None
        if not hasattr(actor, "GetMapper"):
            logger.debug("This actor does not have a GetMapper method")
        else:
            mapper = actor.GetMapper()

        if mapper:
            mapper_id = reference_id(mapper)
            mapper_instance = serialize(actor, mapper, mapper_id, context, depth + 1)
            if mapper_instance:
                dependencies.append(mapper_instance)
                calls.append(["setMapper", [wrap_id(mapper_id)]])

        prop = None
        if hasattr(actor, "GetProperty"):
            prop = actor.GetProperty()
        else:
            logger.debug("This actor does not have a GetProperty method")

        if prop:
            prop_id = reference_id(prop)
            property_instance = serialize(actor, prop, prop_id, context, depth + 1)
            if property_instance:
                dependencies.append(property_instance)
                calls.append(["setProperty", [wrap_id(prop_id)]])

        # Handle texture if any
        texture = None
        if hasattr(actor, "GetTexture"):
            texture = actor.GetTexture()
        else:
            logger.debug("This actor does not have a GetTexture method")

        if texture:
            texture_id = reference_id(texture)
            texture_instance = serialize(actor, texture, texture_id, context, depth + 1)
            if texture_instance:
                dependencies.append(texture_instance)
                calls.append(["addTexture", [wrap_id(texture_id)]])

    # Apply transform
    if actor.GetUserMatrix():
        user_matrix = [0] * 16
        matrix = actor.GetUserMatrix()
        for i in range(4):
            for j in range(4):
                idx = i + 4 * j
                user_matrix[idx] = matrix.GetElement(i, j)
        add_on["userMatrix"] = user_matrix

    should_serialize = mapper_instance and property_instance
    if get_cached_property(actor_id, "visibility") and not actor_visibility:
        should_serialize = True

    if should_serialize:
        return {
            "parent": reference_id(parent),
            "id": actor_id,
            "type": class_name(actor),
            "properties": cache_properties(
                actor_id,
                context,
                {
                    # vtkProp
                    "visibility": actor_visibility,
                    "pickable": actor.GetPickable(),
                    "dragable": actor.GetDragable(),
                    "useBounds": actor.GetUseBounds(),
                    # vtkProp3D
                    "origin": actor.GetOrigin(),
                    "position": actor.GetPosition(),
                    "scale": actor.GetScale(),
                    "orientation": actor.GetOrientation(),
                    # vtkActor
                    "forceOpaque": actor.GetForceOpaque(),
                    "forceTranslucent": actor.GetForceTranslucent(),
                    # additional properties
                    **add_on,
                },
            ),
            "calls": calls,
            "dependencies": dependencies,
        }

    return None


# -----------------------------------------------------------------------------


def cube_axes_serializer(parent, actor, actor_id, context, depth):
    """
    Possible add-on properties for vtk.js:
        gridLines: True,
        axisLabels: None,
        axisTitlePixelOffset: 35.0,
        axisTextStyle: {
            fontColor: 'white',
            fontStyle: 'normal',
            fontSize: 18,
            fontFamily: 'serif',
        },
        tickLabelPixelOffset: 12.0,
        tickTextStyle: {
            fontColor: 'white',
            fontStyle: 'normal',
            fontSize: 14,
            fontFamily: 'serif',
        },
    """
    axis_labels = ["", "", ""]
    if actor.GetXAxisLabelVisibility():
        axis_labels[0] = actor.GetXTitle()
    if actor.GetYAxisLabelVisibility():
        axis_labels[1] = actor.GetYTitle()
    if actor.GetZAxisLabelVisibility():
        axis_labels[2] = actor.GetZTitle()

    text_color = rgb_float_to_hex(*actor.GetXAxesGridlinesProperty().GetColor())

    dependencies = []
    calls = [
        [
            "setCamera",
            [wrap_id(reference_id(actor.GetCamera()))],
        ]
    ]

    prop = None
    if hasattr(actor, "GetXAxesLinesProperty"):
        prop = actor.GetXAxesLinesProperty()
    else:
        logger.debug("This actor does not have a GetXAxesLinesProperty method")

    if prop:
        prop_id = reference_id(prop)
        property_instance = serialize(actor, prop, prop_id, context, depth + 1)
        if property_instance:
            dependencies.append(property_instance)
            calls.append(["setProperty", [wrap_id(prop_id)]])

    return {
        "parent": reference_id(parent),
        "id": actor_id,
        "type": "vtkCubeAxesActor",
        "properties": cache_properties(
            actor_id,
            context,
            {
                # vtkProp
                "visibility": actor.GetVisibility(),
                "pickable": actor.GetPickable(),
                "dragable": actor.GetDragable(),
                "useBounds": actor.GetUseBounds(),
                # vtkProp3D
                "origin": actor.GetOrigin(),
                "position": actor.GetPosition(),
                "scale": actor.GetScale(),
                # vtkActor
                "forceOpaque": actor.GetForceOpaque(),
                "forceTranslucent": actor.GetForceTranslucent(),
                # vtkCubeAxesActor
                "dataBounds": actor.GetBounds(),
                "faceVisibilityAngle": 8,
                "gridLines": True,
                "axisLabels": axis_labels,
                "axisTitlePixelOffset": 35.0,
                "axisTextStyle": {
                    "fontColor": text_color,
                    "fontStyle": "normal",
                    "fontSize": 18,
                    "fontFamily": "serif",
                },
                "tickLabelPixelOffset": 12.0,
                "tickTextStyle": {
                    "fontColor": text_color,
                    "fontStyle": "normal",
                    "fontSize": 14,
                    "fontFamily": "serif",
                },
            },
        ),
        "calls": calls,
        "dependencies": dependencies,
    }


# -----------------------------------------------------------------------------


def scalar_bar_actor_serializer(parent, actor, actor_id, context, depth):
    dependencies = []
    calls = []
    lut = actor.GetLookupTable()
    if not lut:
        return None

    lut_id = reference_id(lut)
    lut_instance = serialize(actor, lut, lut_id, context, depth + 1)
    if not lut_instance:
        return None

    dependencies.append(lut_instance)
    calls.append(["setScalarsToColors", [wrap_id(lut_id)]])

    prop = None
    if hasattr(actor, "GetProperty"):
        prop = actor.GetProperty()
    else:
        if context.debug_all:
            print("This scalar_bar_actor does not have a GetProperty method")

        if prop:
            prop_id = reference_id(prop)
            property_instance = serialize(actor, prop, prop_id, context, depth + 1)
            if property_instance:
                dependencies.append(property_instance)
                calls.append(["setProperty", [wrap_id(prop_id)]])

    axis_label = actor.GetTitle()
    width = actor.GetWidth()
    height = actor.GetHeight()

    return {
        "parent": reference_id(parent),
        "id": actor_id,
        "type": "vtkScalarBarActor",
        "properties": cache_properties(
            actor_id,
            context,
            {
                # vtkProp
                "visibility": actor.GetVisibility(),
                "pickable": actor.GetPickable(),
                "dragable": actor.GetDragable(),
                "useBounds": actor.GetUseBounds(),
                # vtkActor2D
                # "position": actor.GetPosition(),
                # "position2": actor.GetPosition2(),
                # "width": actor.GetWidth(),
                # "height": actor.GetHeight(),
                # vtkScalarBarActor
                "automated": True,
                "axisLabel": axis_label,
                # 'bar_position': [0, 0],
                # 'bar_size': [0, 0],
                "boxPosition": [0.88, -0.92],
                "boxSize": [width, height],
                "axisTitlePixelOffset": 36.0,
                "axisTextStyle": {
                    "fontColor": rgb_float_to_hex(
                        *actor.GetTitleTextProperty().GetColor()
                    ),
                    "fontStyle": "normal",
                    "fontSize": 18,
                    "fontFamily": "serif",
                },
                "tickLabelPixelOffset": 14.0,
                "tickTextStyle": {
                    "fontColor": rgb_float_to_hex(
                        *actor.GetTitleTextProperty().GetColor()
                    ),
                    "fontStyle": "normal",
                    "fontSize": 14,
                    "fontFamily": "serif",
                },
                "drawNanAnnotation": actor.GetDrawNanAnnotation(),
                "drawBelowRangeSwatch": actor.GetDrawBelowRangeSwatch(),
                "drawAboveRangeSwatch": actor.GetDrawAboveRangeSwatch(),
            },
        ),
        "calls": calls,
        "dependencies": dependencies,
    }


# -----------------------------------------------------------------------------


def axes_actor_serializer(parent, actor, actor_id, context, depth):
    actor_visibility = actor.GetVisibility()

    if not actor_visibility:
        return None

    # C++ extract
    label_show = actor.GetAxisLabels()
    # label_position = actor.GetNormalizedLabelPosition()

    # shaft_length = actor.GetNormalizedShaftLength()
    shaft_type = actor.GetShaftType()  # int [line/cylinder]

    tip_length = actor.GetNormalizedTipLength()
    # tip_type = actor.GetTipType() # int [cone/sphere]

    cone_resolution = actor.GetConeResolution()
    cone_radius = actor.GetConeRadius()

    cylinder_resolution = actor.GetCylinderResolution()
    cylinder_radius = actor.GetCylinderRadius()

    # sphere_radius = actor.GetSphereRadius()
    # sphere_resolution = actor.GetSphereResolution()

    # XYZ...
    # actor.GetXAxisCaptionActor2D()
    # actor.GetXAxisTipProperty()
    # actor.GetXAxisShaftProperty()
    # actor.GetXAxisLabelText()

    # Apply transform
    user_matrix = [1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1]
    if actor.GetUserMatrix():
        matrix = actor.GetUserMatrix()
        matrix.Transpose()
        for i in range(4):
            for j in range(4):
                idx = i + 4 * j
                user_matrix[idx] = matrix.GetElement(j, i)

    return {
        "parent": reference_id(parent),
        "id": actor_id,
        "type": "vtkAxesActor",
        "properties": cache_properties(
            actor_id,
            context,
            {
                # vtkProp
                "visibility": actor_visibility,
                "pickable": actor.GetPickable(),
                "dragable": actor.GetDragable(),
                "useBounds": actor.GetUseBounds(),
                # vtkProp3D
                "origin": actor.GetOrigin(),
                "position": actor.GetPosition(),
                "scale": actor.GetScale(),
                "userMatrix": user_matrix,
                # vtkAxesActor
                "labels": {
                    "show": label_show,
                    "x": actor.GetXAxisLabelText(),
                    "y": actor.GetYAxisLabelText(),
                    "z": actor.GetZAxisLabelText(),
                },
                "config": {
                    "recenter": 0,
                    "tipResolution": cone_resolution,  # 60,
                    "tipRadius": 0.2 * cone_radius,  # 0.1,
                    "tipLength": tip_length[0],  # 0.2,
                    "shaftResolution": cylinder_resolution,  # 60,
                    "shaftRadius": 0.01 if shaft_type else cylinder_radius,  # 0.03,
                    "invert": 0,
                },
                "xAxisColor": list(
                    map(lambda x: int(x * 255), actor.GetXAxisTipProperty().GetColor())
                ),
                "yAxisColor": list(
                    map(lambda x: int(x * 255), actor.GetYAxisTipProperty().GetColor())
                ),
                "zAxisColor": list(
                    map(lambda x: int(x * 255), actor.GetZAxisTipProperty().GetColor())
                ),
            },
        ),
    }



import argparse
import base64
import sys
from pathlib import Path
from typing import TextIO

HTML_VIEWER_PATH = Path(__file__).with_name("static_viewer.html")


def data_to_base64(data: bytes):
    base64Content = base64.b64encode(data)
    return base64Content.decode().replace("\n", "")


def write_html(data: bytes, output: TextIO):
    base64Content = data_to_base64(data)
    with open(HTML_VIEWER_PATH, mode="r", encoding="utf-8") as srcHtml:
        for line in srcHtml:
            if "</body>" in line:
                output.write("<script>\n")
                output.write("var container = document.querySelector('.content');\n")
                output.write('var base64Str = "%s";\n\n' % base64Content)
                output.write("OfflineLocalView.load(container, { base64Str });\n")
                output.write("</script>\n")

            output.write(line)


def embed_data_to_viewer_file(data: bytes, output_file: Path):
    with open(output_file, mode="w", encoding="utf-8") as dstHtml:
        write_html(data, dstHtml)


def main():
    parser = argparse.ArgumentParser(description="HTML exporter for local view data")

    parser.add_argument(
        "--input",
        help="Input data file",
    )

    args, _ = parser.parse_known_args()

    if args.input is None:
        parser.print_help()
        sys.exit(0)

    input_file = Path(args.input)
    if not input_file.exists():
        raise FileNotFoundError(f"Input file {input_file.name} not found.")

    input = Path(input_file)
    output = input.with_name(f"{input.name}.html")

    with open(input, "rb") as data:
        data = data.read()

    embed_data_to_viewer_file(data, output)


if __name__ == "__main__":
    main()



from wslink import register as export_rpc

from .web_protocol import vtkWebProtocol


class vtkWebViewPort(vtkWebProtocol):
    """Basic 3_d Viewport API (Camera + Orientation + CenterOfRotation"""

    @export_rpc("viewport.camera.reset")
    def reset_camera(self, view_id):
        """
        RPC callback to reset camera.
        """
        view = self.get_view(view_id)
        renderer = view.GetRenderers().GetFirstRenderer()
        renderer.ResetCamera()

        self.app.InvalidateCache(view)
        self.app.InvokeEvent("UpdateEvent")

        return str(self.get_global_id(view))

    @export_rpc("viewport.axes.orientation.visibility.update")
    def update_orientation_axes_visibility(self, view_id, show_axis):
        """
        RPC callback to show/hide OrientationAxis.
        """
        view = self.get_view(view_id)
        # FIXME seb: view.OrientationAxesVisibility = (show_axis if 1 else 0);

        self.app.InvalidateCache(view)
        self.app.InvokeEvent("UpdateEvent")

        return str(self.get_global_id(view))

    @export_rpc("viewport.axes.center.visibility.update")
    def update_center_axes_visibility(self, view_id, show_axis):
        """
        RPC callback to show/hide CenterAxesVisibility.
        """
        view = self.get_view(view_id)
        # FIXME seb: view.CenterAxesVisibility = (show_axis if 1 else 0);

        self.app.InvalidateCache(view)
        self.app.InvokeEvent("UpdateEvent")

        return str(self.get_global_id(view))

    @export_rpc("viewport.camera.update")
    def update_camera(self, view_id, focal_point, view_up, position, force_update=True):
        view = self.get_view(view_id)

        camera = view.GetRenderers().GetFirstRenderer().GetActiveCamera()
        camera.SetFocalPoint(focal_point)
        camera.SetViewUp(view_up)
        camera.SetPosition(position)

        if force_update:
            self.app.InvalidateCache(view)
            self.app.InvokeEvent("UpdateEvent")



PROP_CACHE = {}


def get_cached_property(obj_id, prop):
    return PROP_CACHE.get(obj_id, {}).get(prop, None)


def cache_properties(obj_id, ctx, props):
    if not ctx.ignore_last_dependencies:
        prev = PROP_CACHE.get(obj_id)
        PROP_CACHE[obj_id] = props

        if prev is None:
            return props

        delta = {}
        if prev != props:
            for key in props:
                if prev.get(key) != props[key]:
                    delta[key] = props[key]
        return delta

    PROP_CACHE[obj_id] = props
    return props



from .local_rendering import ParaViewWebLocalRendering
from .mouse_handler import ParaViewWebMouseHandler
from .publish_image_delivery import ParaViewWebPublishImageDelivery
from .view_port import ParaViewWebViewPort


__all__ = [
    "ParaViewWebLocalRendering",
    "ParaViewWebMouseHandler",
    "ParaViewWebPublishImageDelivery",
    "ParaViewWebViewPort",
]



from paraview import simple
from wslink import register as export_rpc

from .web_protocol import ParaViewWebProtocol


class ParaViewWebViewPort(ParaViewWebProtocol):
    def __init__(self, scale=1.0, max_width=2560, max_height=1440, **kwargs):
        super().__init__()
        self.scale = scale
        self.max_width = max_width
        self.max_height = max_height

    # RpcName: reset_camera => viewport.camera.reset
    @export_rpc("viewport.camera.reset")
    def reset_camera(self, view_id):
        """
        RPC callback to reset camera.
        """
        view = self.get_view(view_id)
        simple.Render(view)
        simple.ResetCamera(view)
        try:
            view.CenterOfRotation = view.CameraFocalPoint
        except AttributeError:
            pass

        self.app.InvalidateCache(view.SMProxy)
        self.app.InvokeEvent("UpdateEvent")

        return view.GetGlobalIDAsString()

    # RpcName: update_orientation_axes_visibility => viewport.axes.orientation.visibility.update
    @export_rpc("viewport.axes.orientation.visibility.update")
    def update_orientation_axes_visibility(self, view_id, show_axis):
        """
        RPC callback to show/hide OrientationAxis.
        """
        view = self.get_view(view_id)
        view.OrientationAxesVisibility = show_axis if 1 else 0

        self.app.InvalidateCache(view.SMProxy)
        self.app.InvokeEvent("UpdateEvent")

        return view.GetGlobalIDAsString()

    # RpcName: update_center_axes_visibility => viewport.axes.center.visibility.update
    @export_rpc("viewport.axes.center.visibility.update")
    def update_center_axes_visibility(self, view_id, show_axis):
        """
        RPC callback to show/hide CenterAxesVisibility.
        """
        view = self.get_view(view_id)
        view.CenterAxesVisibility = show_axis if 1 else 0

        self.app.InvalidateCache(view.SMProxy)
        self.app.InvokeEvent("UpdateEvent")

        return view.GetGlobalIDAsString()

    # RpcName: update_camera => viewport.camera.update
    @export_rpc("viewport.camera.update")
    def update_camera(self, view_id, focal_point, view_up, position, force_update=True):
        view = self.get_view(view_id)

        view.CameraFocalPoint = focal_point
        view.CameraViewUp = view_up
        view.CameraPosition = position

        if force_update:
            self.app.InvalidateCache(view.SMProxy)
            self.app.InvokeEvent("UpdateEvent")

    @export_rpc("viewport.camera.get")
    def get_camera(self, view_id):
        view = self.get_view(view_id)
        bounds = [-1, 1, -1, 1, -1, 1]

        if view and view.GetClientSideView().GetClassName() == "vtkPVRenderView":
            rr = view.GetClientSideView().GetRenderer()
            bounds = rr.ComputeVisiblePropBounds()

        return {
            "bounds": bounds,
            "center": list(view.CenterOfRotation),
            "focal": list(view.CameraFocalPoint),
            "up": list(view.CameraViewUp),
            "position": list(view.CameraPosition),
        }

    @export_rpc("viewport.size.update")
    def update_size(self, view_id, width, height):
        view = self.get_view(view_id)
        w = width * self.scale
        h = height * self.scale
        if w > self.max_width:
            s = float(self.max_width) / float(w)
            w *= s
            h *= s
        elif h > self.max_height:
            s = float(self.max_height) / float(h)
            w *= s
            h *= s
        view.ViewSize = [int(w), int(h)]
        self.app.InvokeEvent("UpdateEvent")



import logging

from .actors import (
    axes_actor_serializer,
    cube_axes_serializer,
    generic_actor_serializer,
    scalar_bar_actor_serializer,
)
from .data import (
    imagedata_serializer,
    generic_volume_serializer,
    merge_to_polydata_serializer,
    polydata_serializer,
)
from .lights import light_serializer
from .lookup_tables import (
    color_transfer_function_serializer,
    discretizable_color_transfer_function_serializer,
    lookup_table_serializer,
    lookup_table_serializer2,
    pwf_serializer,
)
from .mappers import generic_mapper_serializer, generic_volume_mapper_serializer
from .properties import property_serializer, volume_property_serializer
from .registry import register_instance_serializer, register_js_class
from .render_windows import (
    camera_serializer,
    renderer_serializer,
    render_window_serializer,
)
from .textures import texture_serializer

logger = logging.getLogger(__name__)

CONVERT_LUT = False
SKIP_LIGHT = False


def encode_lut(value=True):
    global CONVERT_LUT
    CONVERT_LUT = value


def skip_light(value=True):
    global SKIP_LIGHT
    SKIP_LIGHT = value


def lookup_table_serializer_selector(*args, **kwargs):
    global CONVERT_LUT
    if CONVERT_LUT:
        return lookup_table_serializer2(*args, **kwargs)
    return lookup_table_serializer(*args, **kwargs)


def initialize_serializers():
    # Define which serializer will be used for which VTK classes
    serializers = {
        # Actors/view_props
        generic_actor_serializer: [
            "vtkActor",
            "vtkOpenGLActor",
            "vtkPVLODActor",
        ],
        # Mappers
        generic_mapper_serializer: [
            "vtkMapper",
            "vtkDataSetMapper",
            "vtkPolyDataMapper",
            "vtkImageDataMapper",
            "vtkOpenGLPolyDataMapper",
            "vtkCompositePolyDataMapper2",
            "vtkCompositePolyDataMapper",
        ],
        # Volume mappers
        generic_volume_mapper_serializer: [
            "vtkVolumeMapper",
            "vtkFixedPointVolumeRayCastMapper",
            "vtkGPUVolumeRayCastMapper",
            "vtkOpenGLGPUVolumeRayCastMapper",
            "vtkSmartVolumeMapper",
        ],
        # Volume/view_props
        generic_volume_serializer: "vtkVolume",
        # Textures
        texture_serializer: [
            "vtkTexture",
            "vtkOpenGLTexture",
        ],
        # Properties
        property_serializer: [
            "vtkProperty",
            "vtkOpenGLProperty",
        ],
        # RenderWindows
        render_window_serializer: [
            "vtkRenderWindow",
            "vtkCocoaRenderWindow",
            "vtkXOpenGLRenderWindow",
            "vtkWin32OpenGLRenderWindow",
            "vtkEGLRenderWindow",
            "vtkOpenVRRenderWindow",
            "vtkOpenXRRenderWindow",
            "vtkGenericOpenGLRenderWindow",
            "vtkOSOpenGLRenderWindow",
            "vtkOpenGLRenderWindow",
            "vtkIOSRenderWindow",
            "vtkExternalOpenGLRenderWindow",
            "vtkOffscreenOpenGLRenderWindow",
        ],
        # LookupTables/TransferFunctions
        # lookup_table_serializer: "vtkLookupTable",
        lookup_table_serializer_selector: "vtkLookupTable",
        discretizable_color_transfer_function_serializer: "vtkPVDiscretizableColorTransferFunction",
        color_transfer_function_serializer: "vtkColorTransferFunction",
        pwf_serializer: "vtkPiecewiseFunction",
        # VolumeProperty
        volume_property_serializer: "vtkVolumeProperty",
        # Datasets
        polydata_serializer: "vtkPolyData",
        imagedata_serializer: [
            "vtkImageData",
            "vtkStructuredPoints",
        ],
        merge_to_polydata_serializer: [
            "vtkMultiBlockDataSet",
            "vtkUnstructuredGrid",
        ],
        # Renderers
        renderer_serializer: [
            "vtkRenderer",
            "vtkOpenGLRenderer",
        ],
        # Lights
        light_serializer: []
        if SKIP_LIGHT
        else [
            "vtkLight",
            "vtkPVLight",
            "vtkOpenGLLight",
        ],
        # Annotations (ScalarBar/Axes)
        cube_axes_serializer: "vtkCubeAxesActor",
        scalar_bar_actor_serializer: "vtkScalarBarActor",
        axes_actor_serializer: "vtkAxesActor",
        # Cameras
        camera_serializer: [
            "vtkCamera",
            "vtkOpenGLCamera",
        ],
    }

    js_classes = {
        "vtkMapper": [
            "vtkCompositePolyDataMapper",
            "vtkCompositePolyDataMapper2",
            "vtkDataSetMapper",
            "vtkOpenGLPolyDataMapper",
            "vtkPolyDataMapper",
        ],
        "vtkVolumeMapper": [
            "vtkFixedPointVolumeRayCastMapper",
            "vtkSmartVolumeMapper",
            "vtkOpenGLGPUVolumeRayCastMapper",
        ],
        "vtkImageData": "vtkStructuredPoints",
    }

    for serializer, names in serializers.items():
        if not isinstance(names, (list, tuple)):
            names = [names]

        for name in names:
            register_instance_serializer(name, serializer)

    for js_class, vtk_classes in js_classes.items():
        if not isinstance(vtk_classes, (list, tuple)):
            vtk_classes = [vtk_classes]

        for vtk_class in vtk_classes:
            register_js_class(vtk_class, js_class)



from wslink import register as export_rpc

from vtkmodules.vtkWebCore import vtkWebInteractionEvent

from .web_protocol import ParaViewWebProtocol


class ParaViewWebMouseHandler(ParaViewWebProtocol):
    def __init__(self, **kwargs):
        super().__init__()
        self.last_action = "up"

    # RpcName: mouse_interaction => viewport.mouse.interaction
    @export_rpc("viewport.mouse.interaction")
    def mouse_interaction(self, event):
        """
        RPC Callback for mouse interactions.
        """
        view = self.get_view(event["view"])

        if hasattr(view, "UseInteractiveRenderingForScreenshots"):
            if event["action"] == "down":
                view.UseInteractiveRenderingForScreenshots = 1
            elif event["action"] == "up":
                view.UseInteractiveRenderingForScreenshots = 0

        buttons = 0
        if event["buttonLeft"]:
            buttons |= vtkWebInteractionEvent.LEFT_BUTTON
        if event["buttonMiddle"]:
            buttons |= vtkWebInteractionEvent.MIDDLE_BUTTON
        if event["buttonRight"]:
            buttons |= vtkWebInteractionEvent.RIGHT_BUTTON

        modifiers = 0
        if event["shiftKey"]:
            modifiers |= vtkWebInteractionEvent.SHIFT_KEY
        if event["ctrlKey"]:
            modifiers |= vtkWebInteractionEvent.CTRL_KEY
        if event["altKey"]:
            modifiers |= vtkWebInteractionEvent.ALT_KEY
        if event["metaKey"]:
            modifiers |= vtkWebInteractionEvent.META_KEY

        pvevent = vtkWebInteractionEvent()
        pvevent.SetButtons(buttons)
        pvevent.SetModifiers(modifiers)
        pvevent.SetX(event["x"])
        pvevent.SetY(event["y"])
        # pvevent.SetKeyCode(event["charCode"])
        ret_val = self.app.HandleInteractionEvent(view.SMProxy, pvevent)
        del pvevent

        if event["action"] == "down" and self.last_action != event["action"]:
            self.app.InvokeEvent("StartInteractionEvent")

        if event["action"] == "up" and self.last_action != event["action"]:
            self.app.InvokeEvent("EndInteractionEvent")

        if ret_val:
            self.app.InvokeEvent("UpdateEvent")

        self.last_action = event["action"]

        return ret_val



import math
import os
import importlib
import sys

vtk_module_name = os.environ.get("VTK_MODULE_NAME", "vtkmodules")
sys.modules["vtk_module"] = importlib.import_module(vtk_module_name)

from vtk_module.vtkWebCore import vtkWebInteractionEvent

from wslink import register as export_rpc

from .web_protocol import vtkWebProtocol


def apply_modifiers(event, interactor):
    interactor.SetShiftKey(1 if event.get("shiftKey") else 0)
    interactor.SetControlKey(1 if event.get("ctrlKey") else 0)
    interactor.SetAltKey(1 if event.get("altKey") else 0)
    # interactor.SetMetaKey(1 if event.get("metaKey") else 0)


class vtkWebMouseHandler(vtkWebProtocol):
    """Handle Mouse interaction on any type of view"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.last_action = "up"

    @export_rpc("viewport.mouse.interaction")
    def mouse_interaction(self, event):
        """
        RPC Callback for mouse interactions.
        """
        view = self.get_view(event["view"])

        buttons = 0
        if event["buttonLeft"]:
            buttons |= vtkWebInteractionEvent.LEFT_BUTTON
        if event["buttonMiddle"]:
            buttons |= vtkWebInteractionEvent.MIDDLE_BUTTON
        if event["buttonRight"]:
            buttons |= vtkWebInteractionEvent.RIGHT_BUTTON

        modifiers = 0
        if event["shiftKey"]:
            modifiers |= vtkWebInteractionEvent.SHIFT_KEY
        if event["ctrlKey"]:
            modifiers |= vtkWebInteractionEvent.CTRL_KEY
        if event["altKey"]:
            modifiers |= vtkWebInteractionEvent.ALT_KEY
        if event["metaKey"]:
            modifiers |= vtkWebInteractionEvent.META_KEY

        pvevent = vtkWebInteractionEvent()
        pvevent.SetButtons(buttons)
        pvevent.SetModifiers(modifiers)
        if "x" in event:
            pvevent.SetX(event["x"])
        if "y" in event:
            pvevent.SetY(event["y"])
        if "scroll" in event:
            pvevent.SetScroll(event["scroll"])
        if event["action"] == "dblclick":
            pvevent.SetRepeatCount(2)
        # pvevent.SetKeyCode(event["charCode"])
        ret_val = self.app.HandleInteractionEvent(view, pvevent)
        del pvevent

        if event["action"] == "down" and self.last_action != event["action"]:
            # Make sure animation registration is only triggered on the first
            # "down" action.
            self.app.InvokeEvent("StartInteractionEvent")

        if event["action"] == "up" and self.last_action != event["action"]:
            self.app.InvokeEvent("EndInteractionEvent")

        if ret_val:
            self.app.InvokeEvent("UpdateEvent")

        self.last_action = event["action"]

        return ret_val

    @export_rpc("viewport.mouse.zoom.wheel")
    def update_zoomFromWheel(self, event):
        render_window = self.get_view(event["view"])

        if not render_window:
            return

        interactor = render_window.GetInteractor()
        apply_modifiers(event, interactor)

        if "x" in event and "y" in event:
            # Set the mouse position, so that if there are multiple
            # renderers, the interactor can figure out which one should
            # be modified.
            view_size = render_window.GetSize()
            pos_x = math.floor(view_size[0] * event["x"] + 0.5)
            pos_y = math.floor(view_size[1] * event["y"] + 0.5)

            interactor.SetEventPosition(pos_x, pos_y)
            interactor.MouseMoveEvent()

        if "Start" in event["type"]:
            self.app.InvokeEvent("StartInteractionEvent")
            # It seems every time a StartInteractionEvent is sent, a
            # mouse wheel event with the same spin is sent afterward. We
            # don't want to zoom twice, so do not perform a zoom on the
            # StartInteractionEvent.
            return

        if "End" in event["type"]:
            self.app.InvokeEvent("EndInteractionEvent")
            # This is done
            return

        spin_y = event.get("spinY", 0)
        direction = "Backward" if spin_y >= 0 else "Forward"
        method = getattr(interactor, f"MouseWheel{direction}Event")

        style = interactor.GetInteractorStyle()
        prev_motion_factor = style.GetMouseWheelMotionFactor()
        style.SetMouseWheelMotionFactor(prev_motion_factor * abs(spin_y))
        try:
            method()
        finally:
            style.SetMouseWheelMotionFactor(prev_motion_factor)



from .registry import class_name
from .utils import reference_id
from .cache import cache_properties


def light_type_to_string(value):
    """
    #define VTK_LIGHT_TYPE_HEADLIGHT    1
    #define VTK_LIGHT_TYPE_CAMERA_LIGHT 2
    #define VTK_LIGHT_TYPE_SCENE_LIGHT  3

    'HeadLight';
    'SceneLight';
    'CameraLight'
    """
    if value == 1:
        return "HeadLight"
    elif value == 2:
        return "CameraLight"

    return "SceneLight"


def light_serializer(parent, instance, obj_id, context, depth):
    return {
        "parent": reference_id(parent),
        "id": obj_id,
        "type": class_name(instance),
        "properties": cache_properties(
            obj_id,
            context,
            {
                # 'specular_color': instance.GetSpecularColor(),
                # 'ambient_color': instance.GetAmbientColor(),
                "switch": instance.GetSwitch(),
                "intensity": instance.GetIntensity(),
                "color": instance.GetDiffuseColor(),
                "position": instance.GetPosition(),
                "focalPoint": instance.GetFocalPoint(),
                "positional": instance.GetPositional(),
                "exponent": instance.GetExponent(),
                "coneAngle": instance.GetConeAngle(),
                "attenuationValues": instance.GetAttenuationValues(),
                "lightType": light_type_to_string(instance.GetLightType()),
                "shadowAttenuation": instance.GetShadowAttenuation(),
            },
        ),
    }



class HybridView:
    def __init__(
        self,
        helper,
        view,
        namespace="view",
        mode="local",
        interactive_ratio=1,
        interactive_quality=60,
        still_ratio=1,
        still_quality=98,
    ):
        self._helper = helper
        self._app = helper._trame_server
        self._view = view
        self._view_id = self._helper.id(view)

        # Image setup
        self.i_ratio = interactive_ratio
        self.i_quality = interactive_quality
        self.s_ratio = still_ratio
        self.s_quality = still_quality

        # init keys
        self.ref_key = namespace
        self.mode_key = f"{namespace}Mode" if namespace else "mode"
        self.scene_key = f"{namespace}Scene" if namespace else "scene"
        self.camera_key = f"{namespace}Camera" if namespace else "camera"
        self.animation_key = f"{namespace}Animate" if namespace else "animate"
        self.id_key = f"{namespace}Id" if namespace else "id"

        # Attach annotations
        self._app.state[self.id_key] = self._view_id
        self._app.state[self.mode_key] = mode
        self._app.trigger(self.camera_key)(self.update_camera)
        self._app.trigger(f"{self.animation_key}Start")(self.start_animation)
        self._app.trigger(f"{self.animation_key}Stop")(self.stop_animation)

    def replace_view(self, new_view):
        self._view = new_view
        self._view_id = self._helper.id(new_view)
        self._app.state[self.id_key] = self._view_id

    def start_animation(self):
        self._app.state[self.mode_key] = "remote"
        self._app.protocol_call(
            "viewport.image.push.quality", self._view_id, self.i_quality, self.i_ratio
        )
        self._app.protocol_call("viewport.image.animation.start", self._view_id)

    def stop_animation(self):
        self._app.protocol_call("viewport.image.animation.stop", self._view_id)
        self._app.protocol_call(
            "viewport.image.push.quality", self._view_id, self.s_quality, self.s_ratio
        )
        self.push_geometry()
        self._app.state[self.mode_key] = "local"

    def push_geometry(self):
        self._app.state[self.scene_key] = self._helper.scene(self._view)

    def push_image(self):
        self._helper.push_image(self._view)

    def update_camera(self, camera=None):
        if camera:
            self._helper.set_camera(self._view, **camera)
            self.push_image()
        else:
            # Need to update local camera
            camera = self._helper.camera(self._view)
            self._app.update(ref=self.ref_key, method="setCamera", args=[camera])

    def release_resources(self):
        self._helper.remove_hybrid_view(self.ref_key)
        self._view = None



import logging

import os
import importlib
import sys

vtk_module_name = os.environ.get("VTK_MODULE_NAME", "vtkmodules")
sys.modules["vtk_module"] = importlib.import_module(vtk_module_name)

from vtk_module.vtkFiltersGeometry import vtkCompositeDataGeometryFilter
from vtk_module.vtkFiltersGeometry import vtkDataSetSurfaceFilter

from .helpers import extract_required_fields, get_array_description
from .registry import class_name
from .serialize import serialize
from .utils import reference_id, wrap_id, get_js_array_type

logger = logging.getLogger(__name__)


def polydata_serializer(
    parent, dataset, dataset_id, context, depth, requested_fields=["Normals", "TCoords"]
):
    if dataset and dataset.GetPoints():
        properties = {}

        # Points
        # Handle coordinate conversion if needed
        convert = {}
        js_point_types = get_js_array_type(dataset.GetPoints().GetData())
        if js_point_types not in ["Float32Array", "Float64Array"]:
            convert["dataType"] = (
                "Float32Array" if "32" in js_point_types else "Float64Array"
            )

        points = get_array_description(
            dataset.GetPoints().GetData(), context, **convert
        )
        points["vtkClass"] = "vtkPoints"
        properties["points"] = points

        # Verts
        if dataset.GetVerts() and dataset.GetVerts().GetData().GetNumberOfTuples() > 0:
            _verts = get_array_description(dataset.GetVerts().GetData(), context)
            properties["verts"] = _verts
            properties["verts"]["vtkClass"] = "vtkCellArray"

        # Lines
        if dataset.GetLines() and dataset.GetLines().GetData().GetNumberOfTuples() > 0:
            _lines = get_array_description(dataset.GetLines().GetData(), context)
            properties["lines"] = _lines
            properties["lines"]["vtkClass"] = "vtkCellArray"

        # Polys
        if dataset.GetPolys() and dataset.GetPolys().GetData().GetNumberOfTuples() > 0:
            _polys = get_array_description(dataset.GetPolys().GetData(), context)
            properties["polys"] = _polys
            properties["polys"]["vtkClass"] = "vtkCellArray"

        # Strips
        if (
            dataset.GetStrips()
            and dataset.GetStrips().GetData().GetNumberOfTuples() > 0
        ):
            _strips = get_array_description(dataset.GetStrips().GetData(), context)
            properties["strips"] = _strips
            properties["strips"]["vtkClass"] = "vtkCellArray"

        # Fields
        properties["fields"] = []
        extract_required_fields(
            properties["fields"], parent, dataset, context, requested_fields
        )

        return {
            "parent": reference_id(parent),
            "id": dataset_id,
            "type": class_name(dataset),
            "properties": properties,
        }

    logger.debug("This dataset has no points!")
    return None


# -----------------------------------------------------------------------------


def merge_to_polydata_serializer(
    parent,
    data_object,
    data_object_id,
    context,
    depth,
    requested_fields=["Normals", "TCoords"],
):
    dataset = None

    if data_object.IsA("vtkCompositeDataSet"):
        gf = vtkCompositeDataGeometryFilter()
        gf.SetInputData(data_object)
        gf.Update()
        dataset = gf.GetOutput()
    elif data_object.IsA("vtkUnstructuredGrid"):
        gf = vtkDataSetSurfaceFilter()
        gf.SetInputData(data_object)
        gf.Update()
        dataset = gf.GetOutput()
    else:
        # FIXME: how are we supposed to get 'mapper' here?
        # dataset = mapper.GetInput()
        dataset = None

    return polydata_serializer(
        parent, dataset, data_object_id, context, depth, requested_fields
    )


# -----------------------------------------------------------------------------


def imagedata_serializer(
    parent, dataset, dataset_id, context, depth, requested_fields=["Normals", "TCoords"]
):
    if hasattr(dataset, "GetDirectionMatrix"):
        direction = [dataset.GetDirectionMatrix().GetElement(0, i) for i in range(9)]
    else:
        direction = [1, 0, 0, 0, 1, 0, 0, 0, 1]

    # Extract dataset fields
    fields = []
    extract_required_fields(fields, parent, dataset, context, "*")

    return {
        "parent": reference_id(parent),
        "id": dataset_id,
        "type": class_name(dataset),
        "properties": {
            "spacing": dataset.GetSpacing(),
            "origin": dataset.GetOrigin(),
            "extent": dataset.GetExtent(),
            "direction": direction,
            "fields": fields,
        },
    }


# -----------------------------------------------------------------------------


def generic_volume_serializer(parent, actor, actor_id, context, depth):
    # This kind of actor has two "children" of interest, a property and a
    # mapper
    actor_visibility = actor.GetVisibility()
    mapper_instance = None
    property_instance = None
    calls = []
    dependencies = []
    add_on = {}

    if actor_visibility:
        mapper = None
        if not hasattr(actor, "GetMapper"):
            logger.debug("This actor does not have a GetMapper method")
        else:
            mapper = actor.GetMapper()

        if mapper:
            mapper_id = reference_id(mapper)
            mapper_instance = serialize(actor, mapper, mapper_id, context, depth + 1)
            if mapper_instance:
                dependencies.append(mapper_instance)
                calls.append(["setMapper", [wrap_id(mapper_id)]])

        prop = None
        if hasattr(actor, "GetProperty"):
            prop = actor.GetProperty()
        else:
            logger.debug("This actor does not have a GetProperty method")

        if prop:
            prop_id = reference_id(prop)
            property_instance = serialize(actor, prop, prop_id, context, depth + 1)
            if property_instance:
                dependencies.append(property_instance)
                calls.append(["setProperty", [wrap_id(prop_id)]])

    # Apply transform
    if actor.GetUserMatrix():
        user_matrix = [0] * 16
        matrix = actor.GetUserMatrix()
        for i in range(4):
            for j in range(4):
                idx = i + 4 * j
                user_matrix[idx] = matrix.GetElement(i, j)
        add_on["userMatrix"] = user_matrix

    if actor_visibility == 0 or (mapper_instance and property_instance):
        return {
            "parent": reference_id(parent),
            "id": actor_id,
            "type": class_name(actor),
            "properties": {
                # vtkProp
                "visibility": actor_visibility,
                "pickable": actor.GetPickable(),
                "dragable": actor.GetDragable(),
                "useBounds": actor.GetUseBounds(),
                # vtkProp3D
                "origin": actor.GetOrigin(),
                "position": actor.GetPosition(),
                "scale": actor.GetScale(),
                # additional properties
                **add_on,
            },
            "calls": calls,
            "dependencies": dependencies,
        }

    return None



import base64
import time

from wslink import register as export_rpc
from wslink import schedule_callback

from .web_protocol import vtkWebProtocol


class vtkWebPublishImageDelivery(vtkWebProtocol):
    """Provide publish-based Image delivery mechanism"""

    def __init__(self, decode=True):
        super().__init__()
        self.tracking_views = {}
        self.last_stale_time = {}
        self.stale_handler_count = {}
        self.delta_stale_time_before_render = 0.5  # 0.5s
        self.stale_count_limit = 10
        self.decode = decode
        self.views_in_animations = []
        self.target_frame_rate = 30.0
        self.min_frame_rate = 12.0
        self.max_frame_rate = 30.0

    def push_render(self, v_id, ignore_animation=False, stale_count=0):
        if v_id not in self.tracking_views:
            return

        if not self.tracking_views[v_id]["enabled"]:
            return

        if not ignore_animation and len(self.views_in_animations) > 0:
            return

        if "originalSize" not in self.tracking_views[v_id]:
            view = self.get_view(v_id)
            self.tracking_views[v_id]["originalSize"] = list(view.GetSize())

        if "ratio" not in self.tracking_views[v_id]:
            self.tracking_views[v_id]["ratio"] = 1

        ratio = self.tracking_views[v_id]["ratio"]
        mtime = self.tracking_views[v_id]["mtime"]
        quality = self.tracking_views[v_id]["quality"]
        size = [
            int((s * ratio) + 0.5) for s in self.tracking_views[v_id]["originalSize"]
        ]

        reply = self.still_render(
            {"view": v_id, "mtime": mtime, "quality": quality, "size": size}
        )
        stale = reply["stale"]
        if reply["image"]:
            # depending on whether the app has encoding enabled:
            if self.decode:
                reply["image"] = base64.standard_b64decode(reply["image"])

            reply["image"] = self.addAttachment(reply["image"])
            reply["format"] = "jpeg"
            # save mtime for next call.
            self.tracking_views[v_id]["mtime"] = reply["mtime"]
            # echo back real ID, instead of -1 for 'active'
            reply["id"] = v_id
            self.publish("viewport.image.push.subscription", reply)
        if stale:
            # The image has not been encoded yet.
            self.last_stale_time[v_id] = time.time()
            if self.stale_handler_count[v_id] == 0:
                self.stale_handler_count[v_id] += 1
                schedule_callback(
                    self.delta_stale_time_before_render,
                    lambda: self.render_stale_image(v_id, stale_count),
                )
        else:
            self.last_stale_time[v_id] = 0

    def render_stale_image(self, v_id, stale_count=0):
        if v_id in self.stale_handler_count and self.stale_handler_count[v_id] > 0:
            self.stale_handler_count[v_id] -= 1

            if self.last_stale_time[v_id] != 0:
                delta = time.time() - self.last_stale_time[v_id]
                # Break on stale_count otherwise linked view will always report to be stale
                # And loop forever
                if (
                    delta >= (self.delta_stale_time_before_render * (stale_count + 1))
                    and stale_count < self.stale_count_limit
                ):
                    self.push_render(v_id, False, stale_count + 1)
                elif delta < self.delta_stale_time_before_render:
                    self.stale_handler_count[v_id] += 1
                    schedule_callback(
                        self.delta_stale_time_before_render - delta + 0.001,
                        lambda: self.render_stale_image(v_id, stale_count),
                    )

    def animate(self):
        if len(self.views_in_animations) == 0:
            return

        next_animate_time = time.time() + 1.0 / self.target_frame_rate
        for v_id in self.views_in_animations:
            self.push_render(v_id, True)

        next_animate_time -= time.time()

        if self.target_frame_rate > self.max_frame_rate:
            self.target_frame_rate = self.max_frame_rate

        if next_animate_time < 0:
            if next_animate_time < -1.0:
                self.target_frame_rate = 1
            if self.target_frame_rate > self.min_frame_rate:
                self.target_frame_rate -= 1.0
            schedule_callback(0.001, lambda: self.animate())
        else:
            if (
                self.target_frame_rate < self.max_frame_rate
                and next_animate_time > 0.005
            ):
                self.target_frame_rate += 1.0
            schedule_callback(next_animate_time, lambda: self.animate())

    @export_rpc("viewport.image.animation.fps.max")
    def set_max_frame_rate(self, fps=30):
        self.max_frame_rate = fps

    @export_rpc("viewport.image.animation.fps.get")
    def get_current_frame_rate(self):
        return self.target_frame_rate

    @export_rpc("viewport.image.animation.start")
    def start_view_animation(self, view_id="-1"):
        s_view = self.get_view(view_id)
        real_view_id = str(self.get_global_id(s_view))

        self.views_in_animations.append(real_view_id)
        if len(self.views_in_animations) == 1:
            self.animate()

    @export_rpc("viewport.image.animation.stop")
    def stop_view_animation(self, view_id="-1"):
        s_view = self.get_view(view_id)
        real_view_id = str(self.get_global_id(s_view))

        if real_view_id in self.views_in_animations:
            self.views_in_animations.remove(real_view_id)

    @export_rpc("viewport.image.push")
    def image_push(self, options):
        s_view = self.get_view(options["view"])
        real_view_id = str(self.get_global_id(s_view))
        # Make sure an image is pushed
        self.app.InvalidateCache(s_view)
        self.push_render(real_view_id)

    # Internal function since the reply[image] is not
    # JSON(serializable) it can not be an RPC one
    def still_render(self, options):
        """
        RPC Callback to render a view and obtain the rendered image.
        """
        begin_time = int(round(time.time() * 1000))
        view = self.get_view(options["view"])
        if not view:
            # The view has been deleted, we can not render it...
            # Clean up old view state
            real_view_id = str(self.get_global_id(view))
            if real_view_id in self.views_in_animations:
                self.views_in_animations.remove(real_view_id)
            if real_view_id in self.tracking_views:
                del self.tracking_views[real_view_id]
            if real_view_id in self.stale_handler_count:
                del self.stale_handler_count[real_view_id]

        size = view.GetSize()[0:2]
        resize = size != options.get("size", size)
        if resize:
            size = options["size"]
            if size[0] > 10 and size[1] > 10:
                view.SetSize(size)
        t = 0
        if options and "mtime" in options:
            t = options["mtime"]
        quality = 100
        if options and "quality" in options:
            quality = options["quality"]
        local_time = 0
        if options and "localTime" in options:
            local_time = options["localTime"]
        reply = {}
        app = self.app
        if t == 0:
            app.InvalidateCache(view)
        if self.decode:
            still_render = app.StillRenderToString
        else:
            still_render = app.StillRenderToBuffer
        reply_image = still_render(view, t, quality)

        # Check that we are getting image size we have set if not wait until we
        # do. The render call will set the actual window size.
        tries = 10
        while resize and list(view.GetSize()) != size and size != [0, 0] and tries > 0:
            app.InvalidateCache(view)
            reply_image = still_render(view, t, quality)
            tries -= 1

        if (
            not resize
            and options
            and ("clearCache" in options)
            and options["clearCache"]
        ):
            app.InvalidateCache(view)
            reply_image = still_render(view, t, quality)

        reply["stale"] = app.GetHasImagesBeingProcessed(view)
        reply["mtime"] = app.GetLastStillRenderToMTime()
        reply["size"] = view.GetSize()[0:2]
        reply["memsize"] = reply_image.GetDataSize() if reply_image else 0
        reply["format"] = "jpeg;base64" if self.decode else "jpeg"
        reply["global_id"] = str(self.get_global_id(view))
        reply["localTime"] = local_time
        if self.decode:
            reply["image"] = reply_image
        else:
            # Convert the vtkUnsignedCharArray into a bytes object, required by Autobahn websockets
            reply["image"] = memoryview(reply_image).tobytes() if reply_image else None

        end_time = int(round(time.time() * 1000))
        reply["workTime"] = end_time - begin_time

        return reply

    @export_rpc("viewport.image.push.observer.add")
    def add_render_observer(self, view_id):
        s_view = self.get_view(view_id)
        if not s_view:
            return {"error": "Unable to get view with id %s" % view_id}

        real_view_id = str(self.get_global_id(s_view))

        if real_view_id not in self.tracking_views:

            def observer_callback(*_, **__):
                return self.push_render(real_view_id)

            def start_callback(*_, **__):
                return self.start_view_animation(real_view_id)

            def stop_callback(*_, **__):
                return self.stop_view_animation(real_view_id)

            tag = self.app.AddObserver("UpdateEvent", observer_callback)
            tag_start = self.app.AddObserver("StartInteractionEvent", start_callback)
            tag_stop = self.app.AddObserver("EndInteractionEvent", stop_callback)
            # TODO do we need self.app.AddObserver('ResetActiveView', reset_active_view())
            self.tracking_views[real_view_id] = {
                "tags": [tag, tag_start, tag_stop],
                "observerCount": 1,
                "mtime": 0,
                "enabled": True,
                "quality": 100,
            }
            self.stale_handler_count[real_view_id] = 0
        else:
            # There is an observer on this view already
            self.tracking_views[real_view_id]["observerCount"] += 1

        self.push_render(real_view_id)
        return {"success": True, "viewId": real_view_id}

    @export_rpc("viewport.image.push.observer.remove")
    def remove_render_observer(self, view_id):
        s_view = self.get_view(view_id)
        if not s_view:
            return {"error": "Unable to get view with id %s" % view_id}

        real_view_id = str(self.get_global_id(s_view))

        observer_info = None
        if real_view_id in self.tracking_views:
            observer_info = self.tracking_views[real_view_id]

        if not observer_info:
            return {"error": "Unable to find subscription for view %s" % real_view_id}

        observer_info["observerCount"] -= 1

        if observer_info["observerCount"] <= 0:
            for tag in observer_info["tags"]:
                self.app.RemoveObserver(tag)
            del self.tracking_views[real_view_id]

        return {"result": "success"}

    @export_rpc("viewport.image.push.quality.get")
    def get_view_quality(self, view_id):
        response = dict(quality=1, ratio=1)
        s_view = self.get_view(view_id)

        if s_view:
            real_view_id = str(self.get_global_id(s_view))
            if real_view_id in self.tracking_views:
                observer_info = self.tracking_views[real_view_id]
                response["quality"] = observer_info.get("quality", 100)
                response["ratio"] = observer_info.get("ratio", 1)

        return response

    @export_rpc("viewport.image.push.quality")
    def set_view_quality(self, view_id, quality, ratio=1):
        s_view = self.get_view(view_id)
        if not s_view:
            return {"error": "Unable to get view with id %s" % view_id}

        real_view_id = str(self.get_global_id(s_view))
        observer_info = None
        if real_view_id in self.tracking_views:
            observer_info = self.tracking_views[real_view_id]

        if not observer_info:
            return {"error": "Unable to find subscription for view %s" % real_view_id}

        observer_info["quality"] = quality
        observer_info["ratio"] = ratio

        # Update image size right now!
        if "originalSize" in self.tracking_views[real_view_id]:
            size = [
                int((s * ratio) + 0.5)
                for s in self.tracking_views[real_view_id]["originalSize"]
            ]
            if hasattr(s_view, "SetSize"):
                s_view.SetSize(size)
            else:
                s_view.ViewSize = size

        return {"result": "success"}

    @export_rpc("viewport.image.push.original.size")
    def set_view_size(self, view_id, width, height):
        s_view = self.get_view(view_id)
        if not s_view:
            return {"error": "Unable to get view with id %s" % view_id}

        real_view_id = str(self.get_global_id(s_view))
        observer_info = None
        if real_view_id in self.tracking_views:
            observer_info = self.tracking_views[real_view_id]

        if not observer_info:
            return {"error": "Unable to find subscription for view %s" % real_view_id}

        observer_info["originalSize"] = [width, height]

        return {"result": "success"}

    @export_rpc("viewport.image.push.enabled")
    def enable_view(self, view_id, enabled):
        s_view = self.get_view(view_id)
        if not s_view:
            return {"error": "Unable to get view with id %s" % view_id}

        real_view_id = str(self.get_global_id(s_view))
        observer_info = None
        if real_view_id in self.tracking_views:
            observer_info = self.tracking_views[real_view_id]

        if not observer_info:
            return {"error": "Unable to find subscription for view %s" % real_view_id}

        observer_info["enabled"] = enabled

        return {"result": "success"}

    @export_rpc("viewport.image.push.invalidate.cache")
    def invalidate_cache(self, view_id):
        s_view = self.get_view(view_id)
        if not s_view:
            return {"error": "Unable to get view with id %s" % view_id}

        self.app.InvalidateCache(s_view)
        self.app.InvokeEvent("UpdateEvent")
        return {"result": "success"}



from pathlib import Path

from trame_vtk import __version__


# Compute local path to serve
serve_path = str(Path(__file__).with_name("serve").resolve())
serve_directory = f"__trame_vtk_{__version__}"

# Serve directory for JS/CSS files
serve = {serve_directory: serve_path}

# List of JS files to load (usually from the serve path above)
scripts = [f"{serve_directory}/trame-vtk.js"]

# List of Vue plugins to install/load
vue_use = ["vue_vtk"]



from .registry import class_name
from .serialize import serialize
from .utils import reference_id, wrap_id


def renderer_serializer(parent, instance, obj_id, context, depth):
    dependencies = []
    view_prop_ids = []
    lights_ids = []
    calls = []

    # Camera
    camera = instance.GetActiveCamera()
    camera_id = reference_id(camera)
    camera_instance = serialize(instance, camera, camera_id, context, depth + 1)
    if camera_instance:
        dependencies.append(camera_instance)
        calls.append(["setActiveCamera", [wrap_id(camera_id)]])

    # View prop as representation containers
    view_prop_collection = instance.GetViewProps()
    for rp_idx in range(view_prop_collection.GetNumberOfItems()):
        view_prop = view_prop_collection.GetItemAsObject(rp_idx)
        view_prop_id = reference_id(view_prop)

        view_prop_instance = serialize(
            instance, view_prop, view_prop_id, context, depth + 1
        )
        if view_prop_instance:
            dependencies.append(view_prop_instance)
            view_prop_ids.append(view_prop_id)

    calls += context.build_dependency_call_list(
        "%s-props" % obj_id, view_prop_ids, "addViewProp", "removeViewProp"
    )

    # Lights
    light_collection = instance.GetLights()
    for light_idx in range(light_collection.GetNumberOfItems()):
        light = light_collection.GetItemAsObject(light_idx)
        light_id = reference_id(light)

        light_instance = serialize(instance, light, light_id, context, depth + 1)
        if light_instance:
            dependencies.append(light_instance)
            lights_ids.append(light_id)

    calls += context.build_dependency_call_list(
        "%s-lights" % obj_id, lights_ids, "addLight", "removeLight"
    )

    if len(dependencies) > 1:
        return {
            "parent": reference_id(parent),
            "id": obj_id,
            "type": class_name(instance),
            "properties": {
                "background": instance.GetBackground(),
                "background2": instance.GetBackground2(),
                "viewport": instance.GetViewport(),
                # These commented properties do not yet have real setters in vtk.js
                # 'gradient_background': instance.GetGradientBackground(),
                # 'aspect': instance.GetAspect(),
                # 'pixel_aspect': instance.GetPixelAspect(),
                # 'ambient': instance.GetAmbient(),
                "twoSidedLighting": instance.GetTwoSidedLighting(),
                "lightFollowCamera": instance.GetLightFollowCamera(),
                "layer": instance.GetLayer(),
                "preserveColorBuffer": instance.GetPreserveColorBuffer(),
                "preserveDepthBuffer": instance.GetPreserveDepthBuffer(),
                "nearClippingPlaneTolerance": instance.GetNearClippingPlaneTolerance(),
                "clippingRangeExpansion": instance.GetClippingRangeExpansion(),
                "useShadows": instance.GetUseShadows(),
                "useDepthPeeling": instance.GetUseDepthPeeling(),
                "occlusionRatio": instance.GetOcclusionRatio(),
                "maximumNumberOfPeels": instance.GetMaximumNumberOfPeels(),
                "interactive": instance.GetInteractive(),
            },
            "dependencies": dependencies,
            "calls": calls,
        }

    return None


# -----------------------------------------------------------------------------


def camera_serializer(parent, instance, obj_id, context, depth):
    return {
        "parent": reference_id(parent),
        "id": obj_id,
        "type": class_name(instance),
        "properties": {
            "focalPoint": instance.GetFocalPoint(),
            "position": instance.GetPosition(),
            "viewUp": instance.GetViewUp(),
            "clippingRange": instance.GetClippingRange(),
        },
    }


# -----------------------------------------------------------------------------


def render_window_serializer(parent, instance, obj_id, context, depth):
    dependencies = []
    renderer_ids = []

    renderer_collection = instance.GetRenderers()
    for r_idx in range(renderer_collection.GetNumberOfItems()):
        # Grab the next vtkRenderer
        renderer = renderer_collection.GetItemAsObject(r_idx)
        renderer_id = reference_id(renderer)
        renderer_instance = serialize(
            instance, renderer, renderer_id, context, depth + 1
        )
        if renderer_instance:
            dependencies.append(renderer_instance)
            renderer_ids.append(renderer_id)

    calls = context.build_dependency_call_list(
        obj_id, renderer_ids, "addRenderer", "removeRenderer"
    )

    return {
        "parent": reference_id(parent),
        "id": obj_id,
        "type": class_name(instance),
        "properties": {"numberOfLayers": instance.GetNumberOfLayers()},
        "dependencies": dependencies,
        "calls": calls,
        "mtime": instance.GetMTime(),
    }



import base64
import hashlib
import sys


def rgb_float_to_hex(r, g, b):
    return f"#{int(r*255):02x}{int(g*255):02x}{int(b*255):02x}"


array_types_mapping = [
    " ",  # VTK_VOID                   0
    " ",  # VTK_BIT                    1
    "b",  # VTK_CHAR                   2
    "B",  # VTK_UNSIGNED_CHAR          3
    "h",  # VTK_SHORT                  4
    "H",  # VTK_UNSIGNED_SHORT         5
    "i",  # VTK_INT                    6
    "I",  # VTK_UNSIGNED_INT           7
    "l",  # VTK_LONG                   8
    "L",  # VTK_UNSIGNED_LONG          9
    "f",  # VTK_FLOAT                 10
    "d",  # VTK_DOUBLE                11
    "L",  # VTK_ID_TYPE               12
    " ",  # unspecified               13
    " ",  # unspecified               14
    "b",  # signed_char               15
    "ll",  # VTK_LONG_LONG            16
    "LL",  # VTK_UNSIGNED_LONG_LONG   17
]

javascript_mapping = {
    "b": "Int8Array",
    "B": "Uint8Array",
    "h": "Int16Array",
    "H": "Int16Array",
    "i": "Int32Array",
    "I": "Uint32Array",
    "l": "Int32Array",
    "L": "Uint32Array",
    "f": "Float32Array",
    "d": "Float64Array",
    "ll": "BigInt64Array",
    "LL": "BigUint64Array",
}


def base64_encode(x):
    return base64.b64encode(x).decode("utf-8")


def hash_data_array(data_array):
    if sys.version_info < (3, 9):
        hashed_bit = hashlib.md5(memoryview(data_array)).hexdigest()
    else:
        hashed_bit = hashlib.md5(
            memoryview(data_array), usedforsecurity=False
        ).hexdigest()
    type_code = array_types_mapping[data_array.GetDataType()]
    return "%s_%d%s" % (hashed_bit, data_array.GetSize(), type_code)


def get_js_array_type(data_array):
    return javascript_mapping[array_types_mapping[data_array.GetDataType()]]


def wrap_id(id_str):
    return "instance:${%s}" % id_str


def reference_id(ref):
    if ref:
        try:
            return ref.__this__[1:17]
        except Exception:
            id_str = str(ref)[-12:-1]
            # print('====> fallback ID %s for %s' % (id_str, ref))
            return id_str
    return "0x0"



import os
import importlib
import sys

vtk_module_name = os.environ.get("VTK_MODULE_NAME", "vtkmodules")
sys.modules["vtk_module"] = importlib.import_module(vtk_module_name)

from vtk_module.vtkRenderingCore import vtkColorTransferFunction

from .helpers import data_table_to_list, linspace
from .registry import class_name
from .utils import reference_id
from .cache import cache_properties


def lookup_table_serializer(parent, lookup_table, lookup_table_id, context, depth):
    # No children in this case, so no additions to bindings and return empty list
    # But we do need to add instance

    lookup_table_range = lookup_table.GetRange()

    lookup_table_hue_range = [0.5, 0]
    if hasattr(lookup_table, "GetHueRange"):
        try:
            lookup_table.GetHueRange(lookup_table_hue_range)
        except Exception:
            pass

    lut_sat_range = lookup_table.GetSaturationRange()
    table = data_table_to_list(lookup_table.GetTable())
    properties = {
        "numberOfColors": lookup_table.GetNumberOfColors(),
        "valueRange": lookup_table_range,
        "hueRange": lookup_table_hue_range,
        # 'alpha_range': lut_alpha_range,  # Causes weird rendering artifacts on client
        "saturationRange": lut_sat_range,
        "nanColor": lookup_table.GetNanColor(),
        "belowRangeColor": lookup_table.GetBelowRangeColor(),
        "aboveRangeColor": lookup_table.GetAboveRangeColor(),
        "useAboveRangeColor": True if lookup_table.GetUseAboveRangeColor() else False,
        "useBelowRangeColor": True if lookup_table.GetUseBelowRangeColor() else False,
        "alpha": lookup_table.GetAlpha(),
        "vectorSize": lookup_table.GetVectorSize(),
        "vectorComponent": lookup_table.GetVectorComponent(),
        "vectorMode": lookup_table.GetVectorMode(),
        "indexedLookup": lookup_table.GetIndexedLookup(),
    }

    if table:
        properties["table"] = table

    return {
        "parent": reference_id(parent),
        "id": lookup_table_id,
        "type": class_name(lookup_table),
        "properties": cache_properties(lookup_table_id, context, properties),
    }


# -----------------------------------------------------------------------------


def lookup_table_to_color_transfer_function(lookup_table):
    data_table = lookup_table.GetTable()
    table = data_table_to_list(data_table)

    if not table:
        lookup_table.Build()
        table = data_table_to_list(data_table)

    if table:
        ctf = vtkColorTransferFunction()
        ctf.DeepCopy(lookup_table)  # <== needed to capture vector props

        table_range = lookup_table.GetTableRange()
        points = linspace(*table_range, num=len(table))
        for x, rgba in zip(points, table):
            ctf.AddRGBPoint(x, *[x / 255 for x in rgba[:3]])

        return ctf

    return None


def lookup_table_serializer2(parent, lookup_table, lookup_table_id, context, depth):
    ctf = lookup_table_to_color_transfer_function(lookup_table)
    if ctf:
        return color_transfer_function_serializer(
            parent, ctf, lookup_table_id, context, depth
        )

    return None


def color_transfer_function_serializer(parent, instance, obj_id, context, depth):
    nodes = []
    for i in range(instance.GetSize()):
        # x, r, g, b, midpoint, sharpness
        node = [0, 0, 0, 0, 0, 0]
        instance.GetNodeValue(i, node)
        nodes.append(node)

    discretize = 0
    number_of_values = instance.GetSize()
    if hasattr(instance, "GetDiscretize"):
        discretize = (
            instance.GetDiscretize() if hasattr(instance, "GetDiscretize") else 0
        )
        number_of_values = (
            instance.GetNumberOfValues()
            if hasattr(instance, "GetNumberOfValues")
            else 256
        )

    return {
        "parent": reference_id(parent),
        "id": obj_id,
        "type": class_name(instance),
        "properties": cache_properties(
            obj_id,
            context,
            {
                "clamping": 1 if instance.GetClamping() else 0,
                "colorSpace": instance.GetColorSpace(),
                "hSVWrap": 1 if instance.GetHSVWrap() else 0,
                # 'nan_color': instance.GetNanColor(),                  # Breaks client
                # 'below_range_color': instance.GetBelowRangeColor(),    # Breaks client
                # 'above_range_color': instance.GetAboveRangeColor(),    # Breaks client
                # 'use_above_range_color': 1 if instance.GetUseAboveRangeColor() else 0,
                # 'use_below_range_color': 1 if instance.GetUseBelowRangeColor() else 0,
                "allowDuplicateScalars": 1
                if instance.GetAllowDuplicateScalars()
                else 0,
                "alpha": instance.GetAlpha(),
                "vectorComponent": instance.GetVectorComponent(),
                "vectorSize": instance.GetVectorSize(),
                "vectorMode": instance.GetVectorMode(),
                "indexedLookup": instance.GetIndexedLookup(),
                "nodes": nodes,
                "numberOfValues": number_of_values,
                "discretize": discretize,
            },
        ),
    }


def discretizable_color_transfer_function_serializer(
    parent, instance, obj_id, context, depth
):
    ctf = color_transfer_function_serializer(parent, instance, obj_id, context, depth)
    ctf["properties"]["discretize"] = instance.GetDiscretize()
    ctf["properties"]["numberOfValues"] = instance.GetNumberOfValues()
    return ctf


def pwf_serializer(parent, instance, obj_id, context, depth):
    nodes = []

    for i in range(instance.GetSize()):
        # x, y, midpoint, sharpness
        node = [0, 0, 0, 0]
        instance.GetNodeValue(i, node)
        nodes.append(node)

    return {
        "parent": reference_id(parent),
        "id": obj_id,
        "type": class_name(instance),
        "properties": cache_properties(
            obj_id,
            context,
            {
                "range": list(instance.GetRange()),
                "clamping": instance.GetClamping(),
                "allowDuplicateScalars": instance.GetAllowDuplicateScalars(),
                "nodes": nodes,
            },
        ),
    }



import logging

from ..vtk.core import HybridView
from ..vtk.serializers.mesh import mesh as mesh_vtk

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def has_capabilities(*features):
    pass


class Helper:
    def __init__(self, trame_server):
        self._root_protocol = None
        self._trame_server = trame_server
        self._hybrid_views = {}

        try:  # defer need to paraview to support --www usecase
            from paraview import servermanager
            from paraview.modules.vtkPVClientWeb import vtkPVWebApplication
        except ImportError:
            logger.exception("*** ERROR: ParaView is not available!")
        else:
            self._pv_core = vtkPVWebApplication()
            self._pv_core.SetImageEncoding(0)
            self._servermanager = servermanager
            self._mesh_vtk = mesh_vtk

            # Link our custom protocols initialization
            trame_server.add_protocol_to_configure(self.configure_protocol)

    def has_capabilities(self, *features):
        has_capabilities(*features)

    def id(self, pv_proxy):
        if pv_proxy:
            return pv_proxy.GetGlobalIDAsString()
        return ""

    def object(self, pv_id):
        try:
            pv_id = int(pv_id)
        except ValueError:
            return None
        if pv_id <= 0:
            return None
        return self._servermanager._getPyProxy(
            self._servermanager.ActiveConnection.Session.GetRemoteObject(pv_id)
        )

    def mesh(self, proxy, field_to_keep=None, point_arrays=None, cell_arrays=None):
        proxy.UpdatePipeline()
        source = proxy.GetClientSideObject()
        dataset = source.GetOutput()
        return self._mesh_vtk(
            dataset,
            field_to_keep=field_to_keep,
            point_arrays=point_arrays,
            cell_arrays=cell_arrays,
        )

    def scene(
        self,
        view_proxy,
        reset_camera=False,
        new_state=False,
        widgets=None,
        orientation_axis=0,
        **kwargs,
    ):
        # flush data without requiring a render/picture
        tmp = view_proxy.SuppressRendering
        view_proxy.SuppressRendering = 1
        try:
            view_proxy.StillRender()
        finally:
            view_proxy.SuppressRendering = tmp

        scene_state = self._trame_server.protocol_call(
            "viewport.geometry.view.get.state",
            self.id(view_proxy),
            new_state,
            widgets=widgets,
            orientation_axis=orientation_axis,
        )
        if reset_camera:
            scene_state.setdefault("extra", {})["resetCamera"] = 1
        return scene_state

    def export(self, render_window, widgets=None, orientation_axis=0, **kwargs):
        return self._trame_server.protocol_call(
            "viewport.geometry.view.get.export",
            self.id(render_window),
            widgets=widgets,
            orientation_axis=orientation_axis,
        )

    def push_image(self, view_proxy, reset_camera=False):
        if view_proxy.GetPropertyValue("EnableRenderOnInteraction"):
            view_proxy.EnableRenderOnInteraction = 0

        if reset_camera:
            self._trame_server.protocol_call(
                "viewport.camera.reset", self.id(view_proxy)
            )

        return self._trame_server.protocol_call(
            "viewport.image.push", {"view": self.id(view_proxy)}
        )

    def get_current_image_quality(self, render_window):
        return self._trame_server.protocol_call(
            "viewport.image.push.quality.get", self.id(render_window)
        )

    def set_image_quality(self, render_window, quality, ratio):
        self._trame_server.protocol_call(
            "viewport.image.push.quality",
            self.id(render_window),
            quality,
            ratio,
        )

    def start_animation(self, render_window, fps=30, quality=100, ratio=1):
        self._trame_server.protocol_call("viewport.image.animation.fps.max", fps)
        self.set_image_quality(render_window, quality, ratio)
        self._trame_server.protocol_call(
            "viewport.image.animation.start", self.id(render_window)
        )

    def stop_animation(self, render_window):
        self._trame_server.protocol_call(
            "viewport.image.animation.stop", self.id(render_window)
        )

    def camera(self, view_proxy):
        view_proxy.UpdatePropertyInformation()
        return {
            "focalPoint": list(view_proxy.CameraFocalPoint),
            "parallelProjection": view_proxy.CameraParallelProjection,
            "parallelScale": view_proxy.CameraParallelScale,
            "position": list(view_proxy.CameraPosition),
            "viewAngle": view_proxy.CameraViewAngle,
            "viewUp": list(view_proxy.CameraViewUp),
            "centerOfRotation": list(view_proxy.CenterOfRotation),
        }

    def set_camera(self, view_proxy, **kwargs):
        key_to_attr = {
            "focalPoint": "CameraFocalPoint",
            "parallelProjection": "CameraParallelProjection",
            "parallelScale": "CameraParallelScale",
            "position": "CameraPosition",
            "viewAngle": "CameraViewAngle",
            "viewUp": "CameraViewUp",
            "centerOfRotation": "CenterOfRotation",
        }

        for key, attr in key_to_attr.items():
            if key in kwargs:
                setattr(view_proxy, attr, kwargs[key])

    def view(self, view_proxy, name="view", **kwargs):
        return self.add_hybrid_view(name, view_proxy, **kwargs)

    def configure_protocol(self, protocol):
        self._root_protocol = protocol

        from .protocols import (
            ParaViewWebLocalRendering,
            ParaViewWebMouseHandler,
            ParaViewWebPublishImageDelivery,
            ParaViewWebViewPort,
        )

        # Initialize vtk application helper
        self._root_protocol.setSharedObject("app", self._pv_core)

        # Remote rendering - image delivery
        self._root_protocol.registerLinkProtocol(ParaViewWebMouseHandler())
        self._root_protocol.registerLinkProtocol(ParaViewWebViewPort())
        self._root_protocol.registerLinkProtocol(
            ParaViewWebPublishImageDelivery(decode=False)
        )

        # Remote rendering - geometry delivery
        self._root_protocol.registerLinkProtocol(ParaViewWebLocalRendering())

        # Mimic client interactor on server side
        from .core import apply_default_interaction_settings

        apply_default_interaction_settings()

    def add_hybrid_view(
        self,
        name,
        view,
        mode="local",
        interactive_ratio=1,
        interactive_quality=60,
        still_ratio=1,
        still_quality=98,
        force_replace=False,
        **kwargs,
    ):
        if name in self._hybrid_views:
            if force_replace:
                self._hybrid_views[name].replace_view(view)
            else:
                print(f"A view with name ({name}) is already registered")
                print(" => returning previous one")
            return self._hybrid_views[name]

        view_helper = HybridView(
            self,
            view,
            name,
            mode,
            interactive_ratio,
            interactive_quality,
            still_ratio,
            still_quality,
        )
        self._hybrid_views[name] = view_helper
        return view_helper

    def remove_hybrid_view(self, name):
        if name in self._hybrid_views:
            self._hybrid_views.pop(name)


# -----------------------------------------------------------------------------
# Module advanced initialization
# -----------------------------------------------------------------------------
HELPERS_PER_SERVER = {}


def setup(trame_server, **kwargs):
    global HELPERS_PER_SERVER
    HELPERS_PER_SERVER[trame_server.name] = Helper(trame_server)


def get_helper(server):
    return HELPERS_PER_SERVER.get(server.name)



import logging

from .serialize import serialize
from .utils import reference_id, wrap_id

logger = logging.getLogger(__name__)


def texture_serializer(parent, texture, texture_id, context, depth):
    # This kind of mapper requires us to get 2 items: input data and lookup
    # table
    data_object = None
    data_object_instance = None
    calls = []
    dependencies = []

    if hasattr(texture, "GetInput"):
        data_object = texture.GetInput()
    else:
        logger.debug("This texture does not have GetInput method")

    if data_object:
        data_object_id = "%s-texture" % texture_id
        data_object_instance = serialize(
            texture, data_object, data_object_id, context, depth + 1
        )
        if data_object_instance:
            dependencies.append(data_object_instance)
            calls.append(["setInputData", [wrap_id(data_object_id)]])

    if data_object_instance:
        return {
            "parent": reference_id(parent),
            "id": texture_id,
            "type": "vtkTexture",
            "properties": {
                "interpolate": texture.GetInterpolate(),
                "repeat": texture.GetRepeat(),
                "edgeClamp": texture.GetEdgeClamp(),
            },
            "calls": calls,
            "dependencies": dependencies,
        }

    return None



import logging
import os

from .mesh import mesh
from .initialize import initialize_serializers
from .serialize import serialize, serialize_widget
from .export import extract_array_hash
from .synchronization_context import SynchronizationContext
from .utils import reference_id
from .initialize import encode_lut, skip_light

logger = logging.getLogger(__name__)
# By default, only show critical messages for serializers
logger.setLevel(logging.CRITICAL)

if "TRAME_SERIALIZE_DEBUG" in os.environ:
    # If this environment variable is set, print out all messages
    logger.setLevel(logging.DEBUG)


def configure_serializer(**options):
    skip_light(options.get("skip_light", True))
    encode_lut(options.get("encode_lut", True))


__all__ = [
    "configure_serializer",
    "encode_lut",
    "skip_light",
    "reference_id",
    "initialize_serializers",
    "mesh",
    "serialize",
    "SynchronizationContext",
    "serialize_widget",
    "extract_array_hash",
]
